/* Terraform - (C) 1997-1998 Robert Gasch (Robert_Gasch@peoplesoft.com)
 *  - www.peoplesoft.com/peoplepages/g/robert_gasch/terraform/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#include <string.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <unistd.h>

#include "FileIO.h"
#include "GlobalSanityCheck.h"
#include "GlobalTrace.h"



/*
 *  constructor: give everyting a meaningful default value
 */
FileIO::FileIO (const char *fName)
{
	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, "+++ FileIO\n");

	SanityCheck::bailout ((!fName), "fName==NULL", "FileIO::FileIO");
	SanityCheck::bailout ((!strlen(fName)), "strlen(fName)==0", "FileIO::FileIO");

	d_File = NULL;
	d_fName = strdup (fName);
	d_baseName = NULL;
	d_extension = NULL;
}


/*
 *  destructor: clean up, delete any dyncamically allocated memory
 */
FileIO::~FileIO ()
{
	if (d_fName) 
		free (d_fName);			// C-style alloc by strdup
	if (d_baseName)
		free (d_baseName);		// C-style alloc by strdup
	if (d_extension)
		free (d_extension);		// C-style alloc by strdup

	if (d_File)
		this->close ();

	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, "--- FileIO\n");
}


/*
 *  open: open the file with the specified mode and the file pointer
 */
FILE *FileIO::open (char *mode)
{
	SanityCheck::bailout ((!mode), "mode==NULL", "FileIO::open");
	return (fopen (d_fName, mode));
}


/*
 *  close: close the file descriptor
 */
int FileIO::close ()
{
	if (d_File)
		{
		FILE	*f = d_File;
		d_File = NULL;
		fflush (f);
		return (fclose (f));
		}
	else
		return (-1);
}


/*
 *  getFilePtr: return the file pointer this class is using 
 */
FILE *FileIO::getFilePtr ()
{
	return (d_File);
}


/*
 *  getName: return the file name 
 */
char *FileIO::getName ()
{
	return (d_fName);
}


/*
 *  getExtension: return a pointer to the filename extension or NULL
 */
char *FileIO::getExtension (int include_dot)
{
	char	*last=NULL; 		// last found dot
	int	len; 			// full string length

	// only do this once
	if (d_extension)
		return (d_extension);

	len = strlen (d_fName);

	// find last dot
	for (int i=len-1; i>0; i--)
		if (d_fName[i] == '.')
			{
			last=&d_fName[i];
			i=0;
			}

	if (last)
		{
		if (!include_dot)
			last++;
		d_extension = strdup (last);
		}

	return (d_extension);
}


/*
 *  getBasename: return a pointer to a copy of the file basename
 */
char *FileIO::getBasename ()
{
	char	*dot; 
	int	len;

	// only do this once
	if (d_baseName)
		return (d_baseName);

	// find first dot in name 
	dot = strchr (d_fName, '.');

	if (!dot)
		return (d_fName);

	len = dot - d_fName;
	d_baseName = strdup (d_fName);
	d_baseName[len] = '\0';
	return (d_baseName);
}


/*
 *  exists: check if the file exists
 */
bool FileIO::exists ()
{
	struct stat	stat_info;

	// stat returns -1 on error, 0 on success
	if (stat (d_fName, &stat_info))
		return (false);
	else
		return (true);
}

