/* Terraform - (C) 1997-1998 Robert Gasch (Robert_Gasch@peoplesoft.com)
 *  - www.peoplesoft.com/peoplepages/g/robert_gasch/terraform/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#include "GuiBufferedDrawingArea.h"
#include "GlobalSanityCheck.h"
#include "GlobalTrace.h"


/*
 *  constructor: initialize everything, connect signals
 */
GuiBufferedDrawingArea::GuiBufferedDrawingArea () : d_pixmap (0)
{
	GlobalTrace::trace (GlobalTrace::TRACE_FLOW, "+++ GuiBufferedDrawingArea\n");

	/* Signals used to handle backing pixmap */
	connect_to_method (expose_event, this, &handleExposeEvent);
	connect_to_method (configure_event, this, &handleConfigureEvent);

	/* Event signals */
	//connect_to_method (motion_notify_event, this, &handleMotionNotifyEvent);
	//connect_to_method (button_press_event, this, &handleButtonPressEvent);

	set_events (GDK_EXPOSURE_MASK
		| GDK_LEAVE_NOTIFY_MASK
		| GDK_BUTTON_PRESS_MASK
		| GDK_BUTTON_RELEASE_MASK
		| GDK_POINTER_MOTION_MASK
		| GDK_POINTER_MOTION_HINT_MASK);

	d_syncInterval = 1;
	d_noSyncCount = 0;
	d_GC=GTK_WIDGET (gtkobj())->style->black_gc;
}


/*
 *  destructor: free any resources we're still holding
 */
GuiBufferedDrawingArea::~GuiBufferedDrawingArea ()
{
	gdk_pixmap_unref (d_pixmap);

	GlobalTrace::trace (GlobalTrace::TRACE_FLOW, "--- GuiBufferedDrawingArea\n");
}


/*
 *  clear: clear the entire drawing area 
 */
void GuiBufferedDrawingArea::clear (GdkColor *col)
{
	GdkRectangle update_rect;

	update_rect.x = 0;
	update_rect.y = 0;
	update_rect.width = width();
	update_rect.height = height();

	if (!col)
    		gdk_draw_rectangle (d_pixmap, 
			GTK_WIDGET(gtkobj())->style->white_gc,
			TRUE, 0, 0, this->width(), this->height());
	else
		{
		setColors (col);
    		drawRectangle (0, 0, this->width(), this->height(), TRUE);
		}

	if (d_syncInterval > 0)
		sync (&update_rect);
}


/*
 *  sync: copy the specified area from the offscreen pixmap to the window. 
 *  	  The area specified by update_rect is added to the area which 
 * 	  has not been displayed since the last sync();
 */
void GuiBufferedDrawingArea::sync (GdkRectangle *update_rect)
{
#define SYNC_MAX(a, b)		(a >= b ? a : b)
#define SYNC_MIN(a, b)		(a <= b ? a : b)

	bool		didDraw = FALSE;
	unsigned int	minx, maxx, miny, maxy;
	GdkRectangle	r;

	if (!update_rect)
		{
		r.x = r.y = 0;
		r.width = this->width();
		r.height = this->height();
		draw (&r);

		d_noSyncCount = 0;
		d_sMaxx = d_sMaxy = 0;
		d_sMinx = this->width();
		d_sMiny = this->height();
		return;
		}
	else
		{
		r.x = update_rect->x;
		r.y = update_rect->y;
		r.width = update_rect->width;
		r.height = update_rect->height;
		}

	if (d_syncInterval == 1)
		{
		draw (&r);
		didDraw = TRUE;
		}
	else
	if (d_syncInterval > 1)
		{
		d_noSyncCount++;
		minx = SYNC_MIN (r.x, r.x+r.width);
		maxx = SYNC_MAX (r.x, r.x+r.width);
		miny = SYNC_MIN (r.y, r.y+r.height);
		maxy = SYNC_MAX (r.y, r.y+r.height);

		// figure out if new points expand repaint rect
		if (minx < d_sMinx)
			d_sMinx = minx;
		if (maxx > d_sMaxx)
			d_sMaxx = maxx;
		if (miny < d_sMiny)
			d_sMiny = miny;
		if (maxy > d_sMaxy)
			d_sMaxy = maxy;

		// time to sync
		if (d_noSyncCount == d_syncInterval)
			{
			r.x = d_sMinx;
			r.y = d_sMiny;
			r.width = (d_sMaxx-d_sMinx)+1;
			r.height = (d_sMaxy-d_sMiny)+1;
			draw (&r);
			didDraw = TRUE;

			d_noSyncCount = 0;
			d_sMaxx = d_sMaxy = 0;
			d_sMinx = this->width();
			d_sMiny = this->height();
			}
		}

	if (didDraw)
		{
		// This seemed a good idea but causes excessive 
		// flickering when redrawing in interactive mode. 
		// 
		//Gtk_Main *gtkMain = Gtk_Main::instance();
		//while (gtkMain->events_pending())
		//	gtkMain->iteration ();
		}
}


/*
 *  setSync: set the sync interval then sync the entire window 
 *	0 means that we don't sync
 *	>0 means that we sync after every n draw operations
 */
void GuiBufferedDrawingArea::setSync (int syncInterval)
{
	GdkRectangle	update_rect;

	d_syncInterval = syncInterval;

	if (d_syncInterval < 0)			// force valid range
		d_syncInterval = 0;

	// reset sync window
	d_sMaxx = d_sMaxy = 0;
	d_sMinx = update_rect.width;
	d_sMiny = update_rect.height;
	d_noSyncCount = 0;

	// sync entire window
	update_rect.x = 0;
	update_rect.y = 0;
	update_rect.width = width();
	update_rect.height = height();
	sync (&update_rect);
}


/*
 *  setUsize: set the window's size and create a new offscreen pixmap
 */
void GuiBufferedDrawingArea::setUsize (int x, int y)
{
	if (x != width() || y != height())
		{
		if (d_pixmap)
			gdk_pixmap_unref(d_pixmap);
		d_pixmap = gdk_pixmap_new(get_window(), x, y, -1);
    		gdk_draw_rectangle (d_pixmap, 
			GTK_WIDGET(gtkobj())->style->white_gc,
			TRUE, 0, 0, x, y);
		this->set_usize (x, y);
		}

	this->show ();
}


/*
 *  setColors: set the fore- and background colors
 */
void GuiBufferedDrawingArea::setColors (GdkColor *fg, GdkColor *bg)
{
	if (fg)
		gdk_gc_set_foreground (GTK_WIDGET (gtkobj())->style->black_gc, fg);
	if (bg)
		gdk_gc_set_foreground (GTK_WIDGET (gtkobj())->style->black_gc, bg);
}


/*
 *  handleConfigureEvent: create a new backing pixmap of the appropriate size 
 */
int GuiBufferedDrawingArea::handleConfigureEvent (GdkEventConfigure *event)
{
    	//GtkWidget 	*widget = GTK_WIDGET(gtkobj());
	if (event);			// silence compiler warning 

	if (d_pixmap)
		gdk_pixmap_unref(d_pixmap);

	d_pixmap = gdk_pixmap_new(get_window(),  width(), height(), -1);
    	gdk_draw_rectangle (d_pixmap, 
		GTK_WIDGET(gtkobj())->style->white_gc,
		TRUE, 0, 0, width(), height());

	return TRUE;
}


/*
 *  handleExposeEvent: redraw the exposed area from the backing pixmap
 */
int GuiBufferedDrawingArea::handleExposeEvent (GdkEventExpose *event)
{
	gdk_draw_pixmap(get_window(),
	    GTK_WIDGET (gtkobj())->style->fg_gc[GTK_WIDGET_STATE (GTK_WIDGET(gtkobj()))],
	    d_pixmap,
	    event->area.x, event->area.y,
	    event->area.x, event->area.y,
	    event->area.width, event->area.height);
	return FALSE;
}


/*
 *  drawPoint: draw a Point on the screen
 */
void GuiBufferedDrawingArea::drawPoint (int x, int y)
{
	GdkRectangle update_rect;

	update_rect.x = x;
	update_rect.y = y;
	update_rect.width = 1;
	update_rect.height = 1;
	gdk_draw_point (d_pixmap, 
		GTK_WIDGET (gtkobj())->style->black_gc,
		update_rect.x, update_rect.y);

	if (d_syncInterval > 0)
		sync (&update_rect);
}


/*
 *  drawLine: draw a line on the screen
 */
void GuiBufferedDrawingArea::drawLine (int x1, int y1, int x2, int y2)
{
	GdkRectangle update_rect;

	update_rect.x = (x1<x2 ? x1 : x2);
	update_rect.y = (y1<y2 ? y1 : y2);
	update_rect.width = (x1>x2 ? x1 : x2)-update_rect.x+1;
	update_rect.height = (y1>y2 ? y1 : y2)-update_rect.y+1;
	// gdk_draw_line (GTK_WIDGET (this), 
	gdk_draw_line (d_pixmap, 
		GTK_WIDGET (gtkobj())->style->black_gc, 
		x1, y1, x2, y2);

	if (d_syncInterval > 0)
		sync (&update_rect);
}


/*
 *  drawRectangle: draw a rectangle on the screen
 */
void GuiBufferedDrawingArea::drawRectangle (int x1, int y1, 
					    int width, int height, 
					    bool filled)
{
	GdkRectangle update_rect;

	update_rect.x = x1;
	update_rect.y = y1;
	update_rect.width = width;
	update_rect.height = height;
	gdk_draw_rectangle (d_pixmap,
		GTK_WIDGET (gtkobj())->style->black_gc,
		filled,
		update_rect.x, update_rect.y,
		update_rect.width, update_rect.height);

	if (d_syncInterval > 0)
		sync (&update_rect);
}


/*
 *  drawPolygon: draw a polygon on the screen
 */
void GuiBufferedDrawingArea::drawPolygon (GdkPoint *points, gint npoints, 
					  bool filled)
{
	GdkRectangle 	update_rect;
	GdkPoint	*p;
	int		sx=100000, sy=10000, 		// smallest and 
			bx=-100000, by=-100000;		// biggest x/y values

	for (int i=0; i<npoints; i++)
		{
		p = &(points[i]);

		// find max and min x points
		if (p->x < sx)
			sx = p->x;
		else
		if (p->x > bx)
			bx = p->x;

		// find max and min y points
		if (p->y < sy)
			sy = p->y;
		else
		if (p->y > by)
			by = p->y;
		}

	update_rect.x = sx;
	update_rect.y = sy;
	update_rect.width = bx-sx;
	update_rect.height = by-sy;
	gdk_draw_polygon (d_pixmap,
		GTK_WIDGET (gtkobj())->style->black_gc,
		filled, points, npoints);

	if (d_syncInterval > 0)
		sync (&update_rect);
}


/*
 *  drawPolygon: draw a filled and then a non-filled polygon
 */
void GuiBufferedDrawingArea::drawBorderedPolygon (GdkPoint *points, int npoints, 
				GdkColor *borderColor, GdkColor *polygonColor)
{
	setColors (polygonColor, borderColor);
	drawPolygon (points, npoints, TRUE);
	setColors (borderColor, polygonColor);
	drawPolygon (points, npoints, FALSE);
}


/*
 *  drawText: draw the specified text to the screen
 */
void GuiBufferedDrawingArea::drawText (int x1, int y1, 
					char *textData, GdkFont *textFont)
{
	GdkRectangle update_rect;

	update_rect.x = x1;
	update_rect.y = y1;
	update_rect.width = x1 + gdk_string_width (textFont, textData);
	update_rect.height = y1+12;
	gdk_draw_string (d_pixmap, textFont, 
		GTK_WIDGET (gtkobj())->style->black_gc, 
		update_rect.x, update_rect.y, textData);

	if (d_syncInterval > 0)
		sync (&update_rect);
}


/*
 *  drawPixmap: draw the specified pixmap to the screen 
 */
void GuiBufferedDrawingArea::drawPixmap (GdkPixmap *pixmap, 
					 int xsrc, int ysrc, 
					 int xdest, int ydest, 
					 int width, int height)
{
	GdkRectangle update_rect;

	update_rect.x = xdest;
	update_rect.y = ydest;
	update_rect.width = width;
	update_rect.height = height;
	gdk_draw_pixmap (d_pixmap, GTK_WIDGET (gtkobj())->style->black_gc, 
			pixmap, xsrc, ysrc, xdest, ydest, width, height);
}



/*
 *  handleButtonPressEvent: process a button press
 */
int GuiBufferedDrawingArea::handleButtonPressEvent (GdkEventButton *event)
{
	return TRUE;
}


/*
 *  handleMotionNotifyEvent: process a moving mouse 
 */
int GuiBufferedDrawingArea::handleMotionNotifyEvent (GdkEventMotion *event)
{
	return TRUE;
}


