/* Terraform - (C) 1997-1998 Robert Gasch (Robert_Gasch@peoplesoft.com)
 *  - www.peoplesoft.com/peoplepages/g/robert_gasch/terraform/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "HeightField.h" 
#include "GlobalTrace.h"
#include "GlobalSanityCheck.h"



/*
 *  constructor: initialize everying to a meaningful default value
 */
HeightField::HeightField (PTYPE *hf, int xsize, int ysize, char *name)
{
	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, "+++ HeightField\n");

	if (hf)
		{
		d_hf=hf;
		d_xsize = xsize;
		d_ysize = ysize;
		d_size = xsize * ysize;
		}
	else
		{
		d_hf=NULL;
		d_xsize=d_ysize=d_size=0;
		}

	if (name)
		d_name = name;
	else
		d_name = strdup ("Untitled.tga");

	d_max=d_min=d_avg=0;
	d_sealevel=0.33;
	d_loadedFromFile=FALSE;
	d_isSaved=TRUE;
	if (d_hf)
		gatherStatistics ();
}


/*
 *  destructor: clean up, free allocated memory
 */
HeightField::~HeightField ()
{
	if (d_name)
		free (d_name);
	if (d_hf)
		delete [] d_hf;

	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, "--- HeightField\n");
}


/*
 *  SetSealevel: replace the actual Height Field data
 */
int HeightField::setData (PTYPE *hf, int xsize, int ysize)
{
	SanityCheck::bailout ((!hf), "hf==NULL", "HeightField::setData");

	if (d_hf)
		delete [] d_hf;
	d_hf = hf;
	d_hfBck = NULL;
	d_xsize = xsize;
	d_ysize = ysize;
	d_size = xsize*ysize;
	gatherStatistics ();
	return (0);
}


/*
 *  SetSealevel: set the sealevel
 */
int HeightField::setSealevel (PTYPE sealevel)
{
	d_sealevel = sealevel;
	return (0);
}


/*
 *  backup: make a backup copy of the current height field
 */
void HeightField::backup ()
{
	SanityCheck::bailout ((!d_hf), "d_hf==NULL", "HeightField::backup");

	if (d_hfBck)
		delete [] d_hfBck;
	d_hfBck = new PTYPE[d_size];
	memcpy (d_hfBck, d_hf, sizeof(PTYPE)*d_size);
}


/*
 *  restore: restore a backup copy of the current height field. Doesn't
 *  	     calculate statistics! 
 */
void HeightField::restore ()
{
	SanityCheck::bailout ((!d_hf), "d_hf==NULL", "HeightField::restore");
	SanityCheck::bailout ((!d_hfBck), "d_hfBck==NULL", "HeightField::restore");

	if (!d_hfBck)
		return;

	delete [] d_hf;
	d_hf = d_hfBck;
}


/*
 *  destroyBackup: destroy the backup copy of the current height field and 
 * 	free the memory. 
 */
void HeightField::destroyBackup ()
{
	SanityCheck::bailout ((!d_hf), "d_hf==NULL", "HeightField::restore");
	SanityCheck::bailout ((!d_hfBck), "d_hfBck==NULL", "HeightField::restore");

	if (!d_hfBck)
		return;

	delete [] d_hfBck;
	d_hfBck = NULL;
}


/*
 *  gatherStatistics: gather some general statistics about the Height Field
 */
int HeightField::gatherStatistics ()
{
	float	h=0; 

	SanityCheck::bailout ((!d_hf), "d_hf==NULL", "HeightField::gatherStatistics");

	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, "Surveying Height Field ...\n"); 

	rescale ();

	for (int i=0; i<(int)d_size; i++)
		h+=(double)d_hf[i];
	d_avg=h/d_size;

	return (0);
}


/*
 *  pMinMax: return 2 pointers to the min and max height cells
 */
void HeightField::pMinMax (PTYPE *pfloor, PTYPE *pceil)
{
	*pfloor = *pceil = d_hf[0];

	for (int i=1; i<d_size; i++)
		if (d_hf[i] < *pfloor)
			*pfloor = d_hf[i];
		else
		if (d_hf[i] > *pceil)
			*pceil = d_hf[i];
}



/*
 *  rescale: rescale hf to lie between values floor and ceil inclusive 
 */
int HeightField::rescale (float floor, float ceil)	// default: (0.0, 1.0)
{
	int		i;
	double		cfactor;
	char 		buf[80];

	pMinMax (&d_min, &d_max);

	sprintf (buf, "Rescaling %f, %f ---> ", d_min, d_max);
	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, buf);

	if (d_min == d_max)
		{
		printf ("oops ... Height Field is constant");
		fflush (stdout);
		return (-1);
		}

	if (d_min==floor && d_max==ceil)
		{
		GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, "OK\n");
		return (1);
		}

	cfactor=(double)((ceil-floor)/(d_max-d_min));
	for (i=0; i<d_size; i++)
		d_hf[i] = (d_hf[i]-d_min)*cfactor;

	d_min=floor;
	d_max=ceil;

	sprintf (buf, "%f, %f\n", floor, ceil);
	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, buf);

	return (0);
}



/*
 *  Stat: display the statistics
 */
int HeightField::printStatistics ()
{

	printf ("* * * * * * * * * * * Height Field Statistics * * * * * * * * * * *\n");
	printf ("Height Field		: %s, %dx%d, %d\n", d_name, d_xsize, d_ysize, d_size);
	printf ("Min Max, Avg		: %f, %f, %f\n", d_min, d_max, d_avg);
	fflush (stdout);
	return (0);
}

