/* Terraform - (C) 1997-1998 Robert Gasch (Robert_Gasch@peoplesoft.com)
 *  - www.peoplesoft.com/peoplepages/g/robert_gasch/terraform/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#ifndef _HEIGHT_FIELD_H
#define _HEIGHT_FIELD_H


#include "GlobalDefs.h"

#define PTYPE		float

/*
 *  HeightField: a class to hold the basic info for 1 height field
 */ 
class HeightField
	{
	friend class HeightFieldDraw;
	friend class HeightFieldGenSSynth;
	friend class HeightFieldGenSubdiv;
	friend class HeightFieldModel;
	friend class HeightFieldOps;
	friend class HeightFieldReader;
	friend class HeightFieldSubset;
	friend class HeightFieldWriter;

	public:
				HeightField(PTYPE *_hf_=NULL, int xsize=0, int ysize=0, char *name=NULL);
				~HeightField();

		// inline methods which return data 
		PTYPE		*getData () 		{return d_hf;}
		PTYPE		getMin ()		{return d_min;}
		PTYPE		getMax ()		{return d_max;}
		PTYPE		getSealevel ()		{return d_sealevel;}
		int		getWidth () 		{return d_xsize;}		
		int		getHeight () 		{return d_ysize;}
		int		getSize() 		{return d_size;}
		bool		isLoadedFromFile ()	{return d_loadedFromFile;}
		bool		isSaved ()		{return d_isSaved;}
		void		setSaved (bool saved)	{d_isSaved=saved;}
		char		*getName ()		{return d_name;}

		// inline methods defined in this file 
		PTYPE		El (int x, int y);
		PTYPE		El (int off);
		PTYPE		ElSea (int x, int y);
		PTYPE		Elmod (int x, int y);
		PTYPE		ElmodSea (int x, int y);
		PTYPE		Elclip (int x, int y);
		PTYPE		ElclipSea (int x, int y);
		int		Off2Posx (int x);
		int		Off2Posy (int y);
		int		Pos2Off (int x, int y);

		// other methods 
		void		backup ();
		void		restore ();
		void		destroyBackup ();
		int		setData (PTYPE *hf, int xsize, int ysize);
		int		setSealevel (PTYPE _sealevel_);
		int		gatherStatistics ();
		int		printStatistics ();
		void		setLoadedFromFile (bool lff) {d_loadedFromFile=lff;}

	protected:

		void		pMinMax (PTYPE *pfloor, PTYPE *pceil);
		int		rescale (float floor=0.0, float ceil=1.0);

		PTYPE		*d_hf,			// height field elevation data 
				*d_hfBck,		// backup copy
				d_min, d_max,		// pointers to min and max values
				d_avg,			// average elevation 
				d_sealevel;		// sealevel 
		int		d_xsize, d_ysize, 	// side lengths
				d_size;			// map size 
		char		*d_name;		// height field name
		bool		d_loadedFromFile;
		bool		d_isSaved;
	};


// public inlined methods 

/*
 *  El: return the element at Position x,y
 */ 
inline PTYPE HeightField::El (int px, int py)
{
	return (d_hf[Pos2Off(px, py)]);
}

/*
 *  El: return the element at offset off
 */ 
inline PTYPE HeightField::El (int off)
{
	return (d_hf[off]);
}

/*
 *  ElSea: return the element at offset off
 */ 
inline PTYPE HeightField::ElSea (int px, int py)
{
	return (El(px,py)>d_sealevel ? El(px,py) : d_sealevel);
}


/*
 *  Elmod: return the element at (safe) mod(Position)
 */ 
inline PTYPE HeightField::Elmod (int px, int py)
{
	int 	mx = px%d_xsize,
		my = py%d_ysize;
	
	return ( d_hf[((py>=0)?my:(my+d_ysize)%d_ysize)*d_xsize + 	
	   	      ((px>=0)?mx:(mx+d_xsize)%d_xsize) ] );
}

/*
 *  Elmod: return Elmod if point above sealevel 
 */ 
inline PTYPE HeightField::ElmodSea (int px, int py)
{
	return (Elmod(px,py)>d_sealevel ? Elmod(px,py) : d_sealevel);
}

/*
 *  Elclip: return the point constrained to the actual HF boundries
 */ 
inline PTYPE HeightField::Elclip (int px, int py)
{
	return ( d_hf[((py<0)?0:((py>=d_ysize)?(d_ysize-1):py))*d_xsize + 
		      ((px<0)?0:((px>=d_xsize)?(d_xsize-1):px))] );
}

/*
 *  ElclipSea: return Elclip if point is above sealevel
 */ 
inline PTYPE HeightField::ElclipSea (int px, int py)
{
	return (Elclip(px,py)>d_sealevel ? Elclip(px,py) : d_sealevel);
}

/*
 *  Off2Posx: convert a 1D offset to the x vector of a 2D position
 */ 
inline int HeightField::Off2Posx (int off)
{
	return (off%d_xsize);
}


/*
 *  Off2Posy: convert a 1D offset to the y vector of a 2D position
 */ 
inline int HeightField::Off2Posy (int off)
{
	return ((int)(off/d_xsize));
}


/*
 *  Pos2Off: convert a 2D position to a 1D offset/vector. 
 */ 
inline int HeightField::Pos2Off (int px, int py)
{
	return ((int)(py*d_xsize)+px);
}

#endif // _HEIGHT_FIELD_H
