/* Terraform - (C) 1997-1998 Robert Gasch (Robert_Gasch@peoplesoft.com)
 *  - www.peoplesoft.com/peoplepages/g/robert_gasch/terraform/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/* ******************************************************************** */
/* *** HeightFieldDraw is based on code from John Beale's excellent *** */
/* ************************ hf-lab package. *************************** */
/* ******************************************************************** */


#include <math.h>
#include <stdlib.h>
#include "HeightFieldDraw.h"
#include "GuiColormapBands.h"
#include "GlobalSanityCheck.h"
#include "GlobalTrace.h"


void resize (int x, int y);
static double ATAN2 (double y, double x);


/*
 *  constructor: initialize everything, pass in the main terraform window
 * 		 to draw into.  
 */
HeightFieldDraw::HeightFieldDraw (HeightField 		*HF, 	// Height Field 
				  TFWindow		*tfWin)	// Drawable
{
	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, "+++ HeightFieldDraw\n");

	SanityCheck::bailout ((!HF), "HF==NULL", "HeightFieldDraw::HeightFieldDraw");
	SanityCheck::bailout ((!HF->d_hf), "HF->d_hf==NULL", "HeightFieldDraw::HeightFieldDraw");
	SanityCheck::bailout ((!tfWin), "tfWin==NULL", "HeightFieldDraw::HeightFieldDraw");

	d_HF = HF;
	d_HFM = NULL;
	d_tfWin = tfWin;
	d_drawArea = (dynamic_cast<GuiBufferedDrawingArea*>(d_tfWin));
	d_tfWin->setSync (d_HF->d_xsize*5);

	d_cMap = NULL;
	d_ncolors=0;
	d_nbands=0;
	d_bandsize=0;
	d_bandLandOffset=0;

	d_erasewin=TRUE; 
	d_scale=1;
	d_fastWire=FALSE;

	resetView ();
}


/*
 *  constructor: initialize everything, pass in a buffered drawing area 
 * 		 to draw to. This typically signifies that we're working 
 * 		 on a preview and can't do window ops (tfWin==NULL).  
 */
HeightFieldDraw::HeightFieldDraw (HeightField 		*HF, 	// Height Field 
				  GuiBufferedDrawingArea *da)
{
	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, "+++ HeightFieldDraw\n");

	SanityCheck::bailout ((!HF), "HF==NULL", "HeightFieldDraw::HeightFieldDraw");
	SanityCheck::bailout ((!HF->d_hf), "HF->d_hf==NULL", "HeightFieldDraw::HeightFieldDraw");
	SanityCheck::bailout ((!da), "da==NULL", "HeightFieldDraw::HeightFieldDraw");

	d_HF = HF;
	d_HFM = NULL;
	d_tfWin = NULL;
	d_drawArea = da;
	d_drawArea->setSync (d_HF->d_xsize*5);

	d_cMap = NULL;
	d_ncolors=0;
	d_nbands=0;
	d_bandsize=0;
	d_bandLandOffset=0;

	d_erasewin=TRUE; 
	d_scale=1;
	d_fastWire=FALSE;

	resetView ();
}

/*
 *  destructor: nothing to clean up
 */
HeightFieldDraw::~HeightFieldDraw ()
{
	if (d_HFM)
		delete d_HFM;

	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, "--- HeightFieldDraw\n");
}


/*
 *  setColormap: set the colormap we will use to draw 
 *	the 3 numeric parameters default to 0 and should only be 
 *	used for color band colormaps
 */
int HeightFieldDraw::setColormap (GuiColormap *colorMap, 
				   int numbands, int bsize, int blandoffset) 
{
	SanityCheck::bailout ((!colorMap), "colorMap==NULL", "HeightFieldDraw::setColormap");

	if (d_cMap == colorMap)
		return 1;

	d_cMap = colorMap;
	d_nbands = numbands;
	d_bandsize = bsize;
	d_bandLandOffset = blandoffset;
	
	SanityCheck::bailout ((!d_cMap->d_linear && (!d_nbands || !d_bandsize || !blandoffset)), 
			"non-linear colormap without band specs", 
			"HeightFieldDraw::setColormap");

	d_ncolors=d_cMap->numColors()-1;
	return 0;
}


void HeightFieldDraw::setSync (int syncRate)
{
	d_drawArea->setSync (syncRate);
}


/*
 *  setMode: set the drawing mode
 */
int HeightFieldDraw::setMode (char *s)
{
	int	oldShader = d_shader;
	char 	buf[80];

	SanityCheck::bailout ((!s), "s==NULL", "HeightFieldDraw::setMode");

	/* *** figure out which mode we should set *** */
	if (!strcmp (s, MENU_VIEW_2DPLANE) || !strcmp (s, "0"))
		{
		if (d_shader == WIRE || d_shader == HEIGHT)
			d_wireResolution=d_resolution;
		d_shader=PLANE;
		d_resolution=1;
		}
	else
	if (!strcmp (s, MENU_VIEW_3DWIRE) || !strcmp (s, "1"))
		{
		d_shader=WIRE;
		d_resolution=d_wireResolution;
		}
	else
	if (!strcmp (s, MENU_VIEW_3DHEIGHT) || !strcmp (s, "2"))
		{
		d_shader=HEIGHT;
		d_resolution=d_wireResolution;
		}
	else
	if (!strcmp (s, MENU_VIEW_3DLIGHT) || !strcmp (s, "3"))
		{
		if (d_shader == WIRE || d_shader == HEIGHT)
			d_wireResolution=d_resolution;
		d_shader=LIGHT;
		d_resolution=1;
		}
	else
		{
		cerr << "Error: Invalid mode [" << s << "] supplied ...\n";
		d_shader=PLANE;
		return (-1);
		}

	if (d_shader == oldShader)
		return (1);

	sprintf (buf, "Mode set to %s\n", s);
	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, buf);

	return (0);
}



/*
 *  setScale: set the scale at which we'll look at the height field
 */
int HeightFieldDraw::setScale (float s, int *x, int *y)
{
	SanityCheck::bailout ((!d_tfWin), "d_tfWin==NULL", "HeightFieldDraw::setScale");

	/* *** figure out which mode we should set *** */
	if (s==0.25 || s==0.5 || s==1 || s==2 || s==3 || s==4 ||
	    s==5 || s==6 || s==7 || s==8 || s==9 || s==10)
		{
		if ((d_HF->d_xsize*s <= MAX_SCREEN_SIZE) &&
		    (d_HF->d_ysize*s <= MAX_SCREEN_SIZE) &&
		    (d_HF->d_xsize*s >= MIN_SCREEN_SIZE) &&
		    (d_HF->d_ysize*s >= MIN_SCREEN_SIZE))
			{
			d_xgsize=(int)(d_HF->d_xsize*s);// size of graphics window
			d_ygsize=(int)(d_HF->d_ysize*s);// size of graphics window
			}
		else
			return 1;

		d_scale = s;
		d_tfWin->setSize (d_xgsize, d_ygsize);

		// avoid bombing out the first time around 
		if (d_cMap)
			draw ();

		if (x) *x=d_xgsize;
		if (y) *y=d_ygsize;

		}
	else
		{
		if (x) *x=d_xgsize;
		if (y) *y=d_ygsize;
		cerr << "Error: Invalid scale [" << s << "] supplied ...\n";
		cerr << "Valid scale factors are: 1, 2, 3\n"; 
		return (-1);
		}

	return (TRUE);
}


/*
 *  alterScale: either double or half the current display scale
 */
int HeightFieldDraw::alterScale (float s, int *x, int *y)
{
	SanityCheck::bailout ((!d_tfWin), "d_tfWin==NULL", "HeightFieldDraw::setScale");

	if ((s != 0.5) && (s != 1) && (s != 2))
		return -1;
	
	if ((d_HF->d_xsize*s <= MAX_SCREEN_SIZE) &&
	    (d_HF->d_ysize*s <= MAX_SCREEN_SIZE) &&
	    (d_HF->d_xsize*s >= MIN_SCREEN_SIZE) &&
	    (d_HF->d_ysize*s >= MIN_SCREEN_SIZE) &&
	    (d_scale*s >= MIN_SCALE_FACTOR) && (d_scale*s <= MAX_SCALE_FACTOR))
		{
		d_scale = d_scale*s;
		d_xgsize=(int)(d_HF->d_xsize*d_scale);// size of graphics window
		d_ygsize=(int)(d_HF->d_ysize*d_scale);// size of graphics window
		d_tfWin->setSize (d_xgsize, d_ygsize);

		// avoid bombing out the first time around 
		if (d_cMap)
			draw ();

		if (x) *x=d_xgsize;
		if (y) *y=d_ygsize;
		return 0;
		}
	else
		return 1;
}


/*
 *  setFastWire: set the fast wire mode on or off
 */
void HeightFieldDraw::setFastWire (bool fastWire)
{
	d_fastWire = fastWire;
}


/*
 *  setFastOptions: set the fast wire mode on or off
 */
void HeightFieldDraw::setFastOptions (bool halfWireRes, bool halfYscale, bool doRotZ)
{
	if (halfWireRes)
		d_fastWireResAdj = 2;
	else
		d_fastWireResAdj = 1;

	if (halfYscale)
		d_fastYscaleAdj = 0.5;
	else
		d_fastYscaleAdj = 1;

	d_fastDoRotZ = doRotZ;
	if (d_fastWire)
		draw ();
}




/*
 *  setEyePos: set the eye position
 */
void HeightFieldDraw::setEyePos (double _x_, double _y_, double _z_)
{
	d_eye.x=_x_;
	d_eye.y=_y_;
	d_eye.z=_z_;
}


/*
 *  setLeightPos: set the light position
 */
void HeightFieldDraw::setLightPos (double _x_, double _y_, double _z_)
{
	d_light.x=_x_;
	d_light.y=_y_;
	d_light.z=_z_;
}


/*
 *  setYScale: set the Y scaling factor 
 */
void HeightFieldDraw::setYScale (double yscale)
{
	d_ysf=yscale;
}


/*
 *  setSeaFill: should we draw the sea as filled?
 */
void HeightFieldDraw::setSeaFill (int sfill)
{
	d_sea=sfill;
}


/*
 *  setResolution: set the wireframe d_resolution
 */
void HeightFieldDraw::setResolution (int i)
{
	if (i < 0)
		{
		cerr << "Sanity Check Failed: Resolution (" << i << ") out of range ...\n";
		cerr << "Valid Range is > 0\n";
		exit (-1);
		}
	d_resolution=i;
	if (d_shader == WIRE || d_shader == HEIGHT)
		d_wireResolution=d_resolution;
}


void HeightFieldDraw::resetView ()
{
	SanityCheck::bailout ((!d_HF), "HF==NULL", "HeightFieldDraw::HeightFieldDraw");
	SanityCheck::bailout ((!d_HF->d_hf), "HF->d_hf==NULL", "HeightFieldDraw::HeightFieldDraw");
	SanityCheck::bailout ((!d_tfWin), "tfWin==NULL", "HeightFieldDraw::HeightFieldDraw");

	d_wireResolution=d_resolution=5;       	// low d_resolution == high detail
	setMode(MENU_VIEW_3DWIRE);
	d_fastWireResAdj=2;
	d_fastYscaleAdj=0.5;
	d_fastDoRotZ=FALSE;

	d_xgsize=(int)(d_HF->getWidth()*d_scale);
	d_ygsize=(int)(d_HF->getHeight()*d_scale);

	d_slMin=1.0;		// minimum slope (shading offset)
	d_slScale=1.0;		// slope scale factor (shading scale)
	d_sea=TRUE;

	d_sxmid = d_xgsize/2;  	// middle of screen in screen coords 
	d_symid = d_ygsize/2;
	d_zsf = 0.3;		// z coordinate scale factor 
	d_ysf = 0.3;		// vertical (y) coordinate scale factor 
	d_zeps = 0.001;
	d_eyeinc = 0.8;		// viewpoint 0,0,0 in +z direction 
	//d_eyeinc = 0.1;		// viewpoint 0,0,0 in +z direction 
	d_bright = 0.8;  		// lightsource intensity 
	d_maxcon = FALSE;		// maximum contrast 

	d_light.x=1;		// light source position 
	d_light.y=0;
	d_light.z=-1;

	d_eye.x=0;		// eye position
	d_eye.y=ceil(d_HF->d_max)*-2;
	d_eye.z=-3.5;

	if (d_HFM)
		delete d_HFM;
	d_HFM = new HeightFieldModel (d_HF);
}


/*
 *  clear: clear the window 
 */
void HeightFieldDraw::clear ()
{
	d_drawArea->setColors (&(d_cMap->d_black));
	d_drawArea->drawRectangle (0, 0, d_xgsize, d_ygsize, TRUE);

}


/*
 *  draw: setup and call the appropriate drawing routines
 */
int HeightFieldDraw::draw (bool changedHF)
{
	D		lmag;         	// magnitude of a vector 

	SanityCheck::bailout ((!d_cMap), "d_cMap==NULL", "HeightFieldDraw::draw");
	SanityCheck::bailout ((d_shader>LIGHT), "shade mode out of range", "HeightFieldDraw::draw");

	//SanityCheck::warning (TRUE, "HERE", "HeightFieldDraw::tri3d");

	d_cmin = 0; 
	d_cmax = d_ncolors-1;

	/* *** calculate the positive and negative range *** */
	d_nrange = ABS (d_HF->d_min - d_HF->d_sealevel);
	d_nrange *= 1.01;				// make sure we don't overflow
	d_prange = ABS (d_HF->d_max - d_HF->d_sealevel);
	d_prange *= 1.01;				// when checking max and min

	d_xgsize=(int)(d_HF->d_xsize*d_scale);
	d_ygsize=(int)(d_HF->d_ysize*d_scale);
	if (d_tfWin)
		d_tfWin->setSize (d_xgsize, d_ygsize);

	if (d_erasewin)
		clear ();

	if (d_shader == PLANE) 			// Wireframe 
		drawhf2d ();
	else
	if (d_fastWire && d_shader==WIRE)
		{
		if (changedHF)
			d_HFM->refresh (d_cMap, d_resolution*d_fastWireResAdj, 
					d_ysf*d_fastYscaleAdj, d_sea);
		drawhf3dFast ();
		}
	else
		{
		//if (TRACE_DEBUG_2)
		//	printSettings ();

		/* *** assign class variables used in further calculations *** */
		lmag = sqrt(d_light.x*d_light.x + d_light.y*d_light.y + d_light.z*d_light.z);
		d_light.x /= lmag;
		d_light.y /= lmag;
		d_light.z /= lmag;

		d_zcenter = -d_eye.z;
		d_ycorr = 2*d_eye.y/(d_zcenter);
		d_sxmid = d_xgsize / 2;  		// middle of screen in screen coords 
		d_symid = d_ygsize / 2;
		d_yoff = (int) (d_symid*(d_ycorr + d_eye.y/d_zcenter)); 

		/* *** do the actual drawing *** */
		drawhf3d ();
		}

	d_drawArea->setColors (&(d_cMap->d_black));
	d_drawArea->sync ();
	gdk_flush ();
	return(0);
}



/* ********************************************************************* */
/* ******************* execute the command in token ******************** */
/* ********************************************************************* */

/*
int HeightFieldDraw::P_Execute ()
{
	int	redraw=TRUE;
	char	*c=P_Tok(0);

	while (*c)
		{
		if (*c == '+') 				// y scaling factor
			d_ysf *= 1.1;
		else if (*c == '-' || *c == '_') 
			ysf /= 1.1;
		else if (*c == 'r') 			// x axis (right, left)
			d_eye.x -= d_eyeinc;
		else if (*c == 'l') 
			eye.x += d_eyeinc;
		else if (*c == 'u') 			// y axis (up, down)
			eye.y -= d_eyeinc;
		else if (*c == 'd') 
			eye.y += d_eyeinc;
		else if (*c == 'b') 			// z axis (back, forward)
			eye.z -= d_eyeinc;
		else if (*c == 'f') 
			eye.z += d_eyeinc;
		else if (*c == '>') 			// brightness 
			bright *= 1.1;
		else if (*c == '<') 
			bright /= 1.1;
		else if (*c == '*') 			// scale 
			{ 
			d_xgsize*=d_scalefact; 
			d_ygsize*=d_scalefact; 
			X_WResize (d_xgsize, d_ygsize);
			redraw=FALSE;
			}
		else if (*c == '/') 
			{ 
			d_xgsize/=d_scalefact; 
			d_ygsize/=d_scalefact;
			X_WResize (d_xgsize, d_ygsize);
			redraw=FALSE;
			}
		else if (*c == 'm') 
			d_maxcon = !d_maxcon;
		else if (!P_TokArgCmp ("verbose", 1)) 
			{
			d_HF->verbose=!d_HF->verbose;
			redraw=FALSE;
			}
		else if (*c == 'l') 
			{
			double 	lmag;
			d_maxcon = FALSE;      // reset contrast limits 
			sscanf(++c ,"%lf,%lf,%lf",&light.x,&light.y,&light.z);
			lmag = sqrt(light.x*light.x + light.y*light.y + light.z*light.z);
			light.x /= lmag;
			light.y /= lmag;
			light.z /= lmag;
			}
		else if (*c == 'e') 
			{
			d_maxcon = FALSE;      // reset eye position 
			sscanf(++c,"%lf,%lf,%lf",&eye.x,&eye.y,&eye.z);
			if (eye.z > -1.01) 
				eye.z = -1.01;
			}
		else if (*c == 'f') 
			{
			eye.z += 0.5;
			if (eye.z >= -1.1) 
				eye.z = -1.1;
			}
		else if ((*c == 'h') || (*c == '?')) 
			{
			help();
			redraw=FALSE;
			}
		else if (*c == 's') 
			{
			d_maxcon = FALSE;
			if (setMode (++c))
				redraw=FALSE;
			break;
			}
		else if (*c == 'S')
			{
			d_sea=!d_sea;
			}
		else if (*c == 'p') 
			{
			sscanf (++c, "%d", &d_resolution);
			//d_resolution = (int)atol((const char*) *(++c));
			if (d_resolution < 1) 
				d_resolution = 1;
			}
		else if (*c == 'R') ;		// Redraw
		else if (*c == 'v')
			{
			//print_vars ();
			d_HF->verbose=!d_HF->verbose;
			redraw=FALSE;
			}
		else if (*c == 'q')
			{
			PARSE_DONE=TRUE;
			}
		else if (*c == 'C')
			use_color=!use_color;
		else if (*c == 'c')
			{
			int y;		// needed for Off2Posx()
			X_WClear ();
			for (int i=0; 
				 i<colors->ColorsUsed() && i<X_WGetWidth(); 
				 i++)
				 X_WFillRectangle (d_HF->Off2Posx(i), d_HF->Off2Posy(i), 
					1, X_WGetHeight(), 
					colors->ColorEntry (i));
			redraw=FALSE;
			}
		else 
			{
			printf ("Unrecognized token: [%s][", c);
			redraw=FALSE;
			break;
			}
		c++;
		}

	if (redraw)
		X_WRedraw (this->X_WGetWindow ());
	return (0);
}
*/


/*
 *  rotateFastView: pass on parameters to d_HFM
 */
int HeightFieldDraw::rotateFastView (float a, float b, float c)
{
	SanityCheck::bailout ((!d_HFM), "d_HFM==NULL", "HeightFieldDraw::rotateFastView");

	if (d_fastWire)
		{
		d_HFM->transform (a, b, c);
		d_HFM->transformD32 ();
		return 0;
		}	
	return 1;
}


/*
 *  drawHF2d: draw a height field in 2D (top view)
 */
int HeightFieldDraw::drawhf2d()
{
	int			x, y,  		// x, y 
				nLandColors=0,	// number of land colors
				nWaterColors=0,	// number of water colors 
				col;		// color index for a given point 
	float			elvRatio;	// elevation ratio
	GuiColormapBands	*CMB = NULL;

	if (!d_HF->d_hf)
		return (-1);

	if (!d_cMap->d_linear)
		{
		CMB = static_cast<GuiColormapBands*>(d_cMap);
		nWaterColors = d_bandLandOffset*CMB->getBandSize();
		nLandColors = d_ncolors-nWaterColors;
		}

	for (int i=0; i<d_HF->d_size; i++)
		{
		y=d_HF->Off2Posy(i);
		x=d_HF->Off2Posx(i);

		if (d_HF->d_hf[i]>1)
			printf ("%f, %d\n", d_HF->d_hf[i], i);

		if (d_cMap->d_linear)
			col=(int)(d_HF->d_hf[i]*d_ncolors);
		else
			{
			if (d_HF->d_hf[i] <= d_HF->getSealevel())
				{
				elvRatio = d_HF->d_hf[i]/d_nrange;
				col=(int) (elvRatio * nWaterColors);
				}
			else
				{
				elvRatio = (d_HF->d_hf[i]-d_HF->getSealevel())/d_prange;
				col=(int) (elvRatio * nLandColors + nWaterColors);
				}
			}

		d_drawArea->setColors (d_cMap->getColor (col));
		d_drawArea->drawRectangle ((int)(x*d_scale), (int)(y*d_scale), 
					  (int)d_scale, (int)d_scale);
		}

	return (0);
}


/*
 *  drawHF3dFast: draw fast wireframe by drawing precomputed triangles
 */
int HeightFieldDraw::drawhf3dFast ()
{
	int 	lim,
		cIndex,
		width = d_xgsize,		// size of graphics area
		height = d_ygsize,		// size of graphics area
		xsize = d_HF->d_xsize,		// size of HF
		ysize = d_HF->d_ysize;		// size of HF
	PT2	p1, p2, p3;

	d_drawArea->setColors (d_cMap->getColor (d_ncolors-1));
	//d_HFM->transformD32 ();
	lim = d_HFM->getNumTri ();
	for (int i=0; i<lim; i++)
		{
		d_HFM->get2DPoints (i, &p1, &p2, &p3, &cIndex);
		d_drawArea->setColors (d_cMap->getColor (cIndex));
/*
		char	buf[80];
		sprintf (buf, "3DFast: %d\t(%f,%f), (%f,%f), (%f,%f)\n", i, 
			  p1.x, p1.y, p2.x, p2.y, p3.x, p3.y);
		if (GlobalTrace::trace (GlobalTrace::TRACE_DEBUG, buf);
		d_drawArea->drawLine (
			int (width/2 + p1.x*(float(width)/float(xsize))),
			int (height/10 - p2.y*(float(height)/float(ysize))),
			int (width/2 + p2.x*(float(width)/float(xsize))),
			int (height/10 - p2.y*(float(height)/float(ysize))));
*/
		d_drawArea->drawLine (
			int (width/2 + p2.x*(float(width)/float(xsize))),
			int (height/10 - p2.y*(float(height)/float(ysize))),
			int (width/2 + p3.x*(float(width)/float(xsize))),
			int (height/10 - p3.y*(float(height)/float(ysize))));
		d_drawArea->drawLine (
			int (width/2 + p3.x*(float(width)/float(xsize))),
			int (height/10 - p3.y*(float(height)/float(ysize))),
			int (width/2 + p1.x*(float(width)/float(ysize))),
			int (height/10 - p1.y*(float(height)/float(xsize))));
		//if (i==2)
		//	i=lim;
		}
	return 0;
}


/*
 *  drawHF3d: draw a height field in 3D perspective as a series of triangles
 */
int HeightFieldDraw::drawhf3d ()
{
	int 	x,z,izi;
	int 	ymin, ymax, xmin, xmax;
	int 	noffx, noffz;
	D 	xd2, yd2, x0, y0, z0, x1, y1, z1, y2, y3;

	xd2 = d_HF->d_xsize / 2.0;
	yd2 = d_HF->d_ysize / 2.0;
	ymin = 0;
	ymax = d_HF->d_ysize-1;
	xmin = 0;
	xmax = d_HF->d_xsize-1;
	if (!d_maxcon) 
		{
		d_cmax = 0; 
		d_cmin = d_ncolors;	// initialize global lims to extrema 
  		}

	/* *** loop over the two array indices: z = into the screen,  *** */
	/* *** x = to right across the screen, y = vertically upwards *** */
	for (z=ymin; z<ymax; z+=d_resolution) 
		{
		izi = (ymax - z) + ymin;
		if ((z+d_resolution) < ymax) 
			noffz = d_resolution;
		else 
			noffz = ymax-z;
		for (x=xmin; x<xmax; x+=d_resolution) 
			{
			if ((x+d_resolution) < xmax) 
				noffx = d_resolution;
			else 
				noffx = xmax-x;
			x0 = (x-xd2)/xd2; z0 = (izi-yd2)/yd2; 
			x1 = (x+noffx-xd2)/xd2; z1 = (izi-noffz-yd2)/yd2; 
			if (d_sea)
				{
				y0 = (D) d_HF->ElmodSea(x,z);
				y1 = (D) d_HF->ElmodSea(x+noffx,z+noffz);
				y2 = (D) d_HF->ElmodSea(x+noffx,z);
				y3 = (D) d_HF->ElmodSea(x,z+noffz);
				tri3d(x0,z0,x1,z1,y0,y1,y2,y3,d_HF->Pos2Off(x,z)); 
				}
			else
				{
				y0 = (D) d_HF->Elmod(x,z);
				y1 = (D) d_HF->Elmod(x+noffx,z+noffz);
				y2 = (D) d_HF->Elmod(x+noffx,z);
				y3 = (D) d_HF->Elmod(x,z+noffz);
				tri3d(x0,z0,x1,z1,y0,y1,y2,y3,d_HF->Pos2Off(x,z)); 
				}
			// if (d_shader==4 || d_shader==5) { col = (int) (d_ncolors * d_HF->El (x*d_HF->d_xsize/d_xgsize,z*d_HF->d_ysize/d_ygsize)); }
			tri3d(x0,z0,x1,z1,y0,y1,y2,y3,d_HF->Pos2Off(x,z)); 
			} 
		} 

	return (0);
} 


/*
 *  tri3d: project two triangles, given 3D coordinates, on a 2D viewscreen
 *  ===============================================================
 *  project two triangles, given 3D coordinates, on a 2D viewscreen
 *  uses gobal vars:
 *   ysf : vertical heightfield scaling factor
 *   ex,ey,ez : location of eye (viewpoint)
 *   d_symid : y coordinate midpoint
 *   ycorr : y correction factor
 *   yoff  : y offset in screen units 
 *   zcenter : center to 'rotate' appparent hf around
 *   ncolors  : maximum color value 
 *   bright  : lightsource intensity 
 * ===============================================================
 *  shader: 0 = wireframe, 1 = height color, 2 = slope, 3 = lightsource
 */
int HeightFieldDraw::tri3d(D x0, D z0,D x1,D z1,D y0,D y1,D y2,D y3, int offset)
{
	D	zmod0, zmod1, slope1, slope2; 
	int 	c1=0, c2=0, c3=0;		// colors 
	D	nx,ny,nz,nmag;   		// normal vector 
	D	ctheta1, ctheta2;          	// cos(theta)  theta = angle between light and normal */
	PTYPE	elv=d_HF->d_hf[offset];
	int	land_bands=d_nbands-d_bandLandOffset;

	/* three sets of points in 2D -- but must alloc 4 for edge lines */ 
	//PT2	p[4];
	//XPoint p[4];
	GdkPoint p[4];

/*
	printf("t3d %1.3f %1.3f %1.3f %1.3f",x0,z0,x1,z1);
	printf(" %1.3f %1.3f %1.3f %1.3f ",y0,y1,y2,y3);
	printf(" * %d %d ",d_shader,col); 
	fflush(stdout);
*/

	/* *** x0,y0,z0 should be in the range -1..1 		*** */
	/* *** transform to eye coordinate (eye.x,eye.y,eye.z) 	*** */
	y0 *= -d_ysf; y1 *= -d_ysf; y2 *= -d_ysf; y3 *= -d_ysf;                          
	x0 -= d_eye.x; y0 -= d_eye.y; z0 -= d_eye.z;
	x1 -= d_eye.x; y1 -= d_eye.y; z1 -= d_eye.z;
	y2 -= d_eye.y; y3 -= d_eye.y; 
	if ((z1 <= d_zeps) || (z0 <= d_zeps)) 
		return (1);  			// foreground clipping 


	if (d_shader == WIRE) 			// Wireframe 
		{ 
		if (d_cMap->d_linear)
			{
			// white, we hope. brightest color, anyway 
			c1=(int)(d_ncolors*(elv/(d_HF->d_max-(d_HF->d_min))));
			c2=0;
			}
		else
			{
			if (elv <= d_HF->d_sealevel)
				{
				c1=(int)(ABS((d_HF->d_sealevel-(d_HF->d_sealevel-elv))/d_nrange*(d_bandLandOffset)));
				c1*=d_bandsize;
				}
			else
				{
				c1=(int)(ABS(((elv-d_HF->d_sealevel)/d_prange) * (land_bands)));
				c1*=d_bandsize;
				//c1+=(d_bandsize*(d_bandLandOffset+1));
				c1+=(d_bandsize*(d_bandLandOffset));
				}
			c2=0;
			}
		}
	else
	if (d_shader == HEIGHT) 			// Height, before transform 
		{
		if (d_cMap->d_linear)
			{
			c1=(int)(d_ncolors*(elv/(d_HF->d_max-(d_HF->d_min))));
			c2=0;
			}
		else
			{
			if (elv <= d_HF->d_sealevel)
				{
				c2=(int)(ABS((d_HF->d_sealevel-(d_HF->d_sealevel-elv))/d_nrange*(d_bandLandOffset)));
				c2*=d_bandsize;
				}
			else
				{
				c2=(int)(ABS(((elv-d_HF->d_sealevel)/d_prange) * (land_bands)));
				c2*=d_bandsize;
				//c2+=(d_bandsize*(d_bandLandOffset+1));
				c2+=(d_bandsize*(d_bandLandOffset));
				}
			c1=c2+d_bandsize/2;
			}
		} 
	else 
	if (d_shader == LIGHT) 			// compute normal, dot product 
		{				// with lightsource 
		nx = (z0-z1)*(y2-y0);
		ny = (z1-z0)*(x1-x0);
		nz = (x1-x0)*(y2-y1);
		nmag = sqrt(nx*nx + ny*ny + nz*nz);
		/* printf("nxyz %1.3f %1.3f %1.3f  nmag %1.3f ",nx,ny,nz,nmag); fflush(stdout); */
		if (nmag==0) 
			ctheta1 = 0;
		else 
			ctheta1 = (nx*d_light.x + ny*d_light.y + nz*d_light.z) / nmag;

		nx = -(z0-z1)*(y3-y1);
		/* ny = (z1-z0)*(x1-x0); */   /* duplicates line above */
		nz = (x0-x1)*(y3-y0);
		nmag = sqrt(nx*nx + ny*ny + nz*nz);
		if (nmag==0) 
			ctheta2 = 0;
		else
			ctheta2 = (nx*d_light.x + ny*d_light.y + nz*d_light.z) / nmag;

		if (d_cMap->d_linear)
			{
			c1 = (int) ((d_ncolors>>1)*(d_bright*(ctheta1 + 1.0)));
			c2 = (int) ((d_ncolors>>1)*(d_bright*(ctheta2 + 1.0)));
			}
		else
			{
			float cbi; 	// color band index, color band offset

			if (ctheta1 < 0)
				ctheta1 = 0;
			if (ctheta2 < 0)
				ctheta2 = 0;

			if (elv <= d_HF->d_sealevel)
				{
				cbi=(int)(ABS(((d_HF->d_sealevel-(d_HF->d_sealevel-elv))/d_nrange*(d_bandLandOffset))));
				cbi*=d_bandsize;
				}
			else
				{
				cbi=(int)(ABS(((elv-d_HF->d_sealevel)/d_prange) * (land_bands)));
				cbi*=d_bandsize;
				cbi+=(d_bandsize*(d_bandLandOffset));
				}
			c1=(int)(cbi+d_bandsize*ctheta1);
			c2=(int)(cbi+d_bandsize*ctheta2);

			if (c1 < 0 || c2 < 0)
				{
				printf ("C1 = %d, C2 = %d\n", c1, c2);
				printf ("cbi = %d\n", (int)(ABS(((d_HF->d_sealevel-(d_HF->d_sealevel-elv))/d_nrange*(d_bandLandOffset)))));
				printf ("ctheta1 = %f, ctheta2 = %f\n", ctheta1, ctheta2);
				exit (0);
				}
			}
		}

	if (d_maxcon &&(d_cmax!=d_cmin)) 		// if maximize_contrast is TRUE... 
		{
/*
		c1 = (int) (((D)d_ncolors/(d_cmax-d_cmin))*(c1-d_cmin));
		c2 = (int) (((D)d_ncolors/(d_cmax-d_cmin))*(c2-d_cmin));
*/
		c1 = (int) (((D)d_ncolors/d_cmax)*(c1));
		c2 = (int) (((D)d_ncolors/d_cmax)*(c2));
		} 

	char 	t[80];

	sprintf (t, "Colormap Lookup RangeError: no color %d in c1", c1);
	if (SanityCheck::warning ((c1 < 0 || c1 > d_ncolors), 
			t, "HeightFieldGrid::trid3d"))
		{
		if (c1 < 0)
			c1 = 0;
		else 
			c1 = d_ncolors-1;
		}

	sprintf (t, "Colormap Lookup RangeError: no color %d in c2", c2);
	if (SanityCheck::warning ((c2 < 0 || c2 > d_ncolors), 
			t, "HeightFieldGrid::trid3d"))
		{
		if (c2 < 0)
			c2 = 0;
		else 
			c2 = d_ncolors-1;
		}

	if (!d_maxcon) 
		{
		if (c1 > d_cmax) 
			d_cmax = c1;
		if (c1 < d_cmin) 
			d_cmin = c1;
		}

	zmod0 = z0 * d_zsf; 		// z1 multiplied by z scale factor
	zmod1 = z1 * d_zsf;

	p[0].x = (int) (d_sxmid*x0/zmod0) + d_sxmid;  	// x,y,z 0 
	p[0].y = (int) (d_symid*y0/zmod0) + d_symid + d_yoff;
 
	p[1].x = (int) (d_sxmid*x1/zmod1) + d_sxmid;  	// x,y,z 1 
	p[1].y = (int) (d_symid*y1/zmod1) + d_symid + d_yoff;

	p[2].x = (int) (d_sxmid*x1/zmod0) + d_sxmid; 	// x,y,z 2 
	p[2].y = (int) (d_symid*y2/zmod0) + d_symid + d_yoff;

	/* *** we're off the sceen. no need to do anything else *** */
	if ( ((p[0].x<0) && (p[1].x<0) && (p[2].x<0)) ||
	     ((p[0].x>d_xgsize) && (p[1].x>d_xgsize) && (p[2].x>d_xgsize)) ||
	     ((p[0].y<0) && (p[1].y<0) && (p[2].y<0)) ||
	     ((p[0].y>d_ygsize) && (p[1].y>d_ygsize) && (p[2].y>d_ygsize))  )
		return (1);
		
	if ( ((p[0].x==p[1].x) && (p[0].y==p[1].y)) ||
	     ((p[0].x==p[2].x) && (p[0].y==p[2].y)) ||
	     ((p[1].x==p[2].x) && (p[1].y==p[2].y)) )  
		{

		//gdk_draw_points (da->window, da->style->black_gc, p, 1);
		d_drawArea->drawPoint (p[0].x, p[0].y);

		}

	else 
		{
		if (d_shader == HEIGHT)
			{
			d_drawArea->drawBorderedPolygon (p, 3, 
					d_cMap->getColor(c2), 
					d_cMap->getColor(c1));

			}
		else if (d_shader == WIRE) 
			{
			if (d_cMap->d_linear)
				c3=(int)(d_ncolors*(elv/(d_HF->d_max-(d_HF->d_min))));
			else
				c3=c1;
			d_drawArea->setColors (d_cMap->getColor (c3));
			d_drawArea->drawPolygon (p, 3, FALSE);
			} 
		else 
			{
			d_drawArea->drawBorderedPolygon (p, 3, 
					d_cMap->getColor(c1), 
					d_cMap->getColor(c1));
			
			}
		}

	p[1].x = (int) (d_sxmid*x0/zmod0) + d_sxmid;	// x,y,z 0 
	p[1].y = (int) (d_symid*y0/zmod0) + d_symid + d_yoff;

	p[0].x = (int) (d_sxmid*x1/zmod1) + d_sxmid;	// x,y,z 1 
	p[0].y = (int) (d_symid*y1/zmod1) + d_symid + d_yoff;

	p[2].x = (int) (d_sxmid*x0/zmod1) + d_sxmid; 	// x,y,z 3 
	p[2].y = (int) (d_symid*y3/zmod1) + d_symid + d_yoff; 

	if( ((p[0].x==p[1].x) && (p[0].y==p[1].y)) ||
	    ((p[0].x==p[2].x) && (p[0].y==p[2].y)) ||
	    ((p[1].x==p[2].x) && (p[1].y==p[2].y)) )  
		{

		// gdk_draw_points (da->window, da->style->black_gc, p, 1);
		d_drawArea->drawPoint (p[0].x, p[0].y);
		}
	else 
		{
		if (d_shader == HEIGHT)  
			{
			d_drawArea->drawBorderedPolygon (p, 3, 
					d_cMap->getColor(c1), 
					d_cMap->getColor(c2));
			}
		else if (d_shader == WIRE) 
			{
			if (d_cMap->d_linear)
				c3=(int)(d_ncolors*(elv/(d_HF->d_max-(d_HF->d_min))));
			else
				c3=c2;
			d_drawArea->setColors (d_cMap->getColor (c1));
			d_drawArea->drawPolygon (p, 3, FALSE);
					//d_cMap->getColor(c3), 
					//d_cMap->getColor(c1));
			} 
		else 
			{
			d_drawArea->drawBorderedPolygon (p, 3, 
					d_cMap->getColor(c2), 
					d_cMap->getColor(c2));
			}
		}

	return (0);
} 


/*
 *  printSettings: print the settings 
 */
void HeightFieldDraw::printSettings ()
{
	printf ("*************** HFD Parameters *************\n");
	printf ("Graph Size		: %d, %d\n", d_xgsize, d_ygsize);
	printf ("Eye Coords		: %f, %f, %f\n", d_eye.x,  d_eye.y, d_eye.z);
	printf ("Light Vector		: %f, %f, %f\n", d_light.x, d_light.y, d_light.z);
	printf ("Scale, Vertical Scale	: %f, %f\n", d_scale, d_ysf);
	printf ("Lightsource Intensity	: %f\n", d_bright);
	printf ("Contrast Maximization	: %s\n", (d_maxcon ? "ON" : "OFF"));
	printf ("Color min, max		: min=%d, max=%d\n", d_cmin, d_cmax);
	printf ("Grid Spacing		: %d\n", d_resolution);
	printf ("Shading Mode		: %d\n", d_shader);
	printf ("Clear			: %s\n", (d_erasewin ? "ON":"OFF"));
	printf ("HF Colors, Bands, C/B	: %d, %d, %d\n", d_ncolors, d_nbands, d_bandsize);
	printf ("Land cols begin@band	: %d\n\n", d_bandLandOffset);
	fflush (stdout);
}



/* 
 * help: print the help for commands into the window 
 */
void HeightFieldDraw::help ()              /* print help message */
{
	printf ("*************** hf3d Help *************\n");
	printf ("You can enter whitespace delimeted commands into the display window\n");
	printf ("Shading: S0=wireframe, S1=height-shading, S2=slope, S3=lightsource\n");
	printf ("Vector to lightsource: Lx,y,z (affects S3 mode only)\n");
	printf ("Eye Position: Ex,y,z\n");
	printf ("Eye Movement: R=right, L=left, U=Up, D=Down, B=back, F=forward (delta=0.1)\n");
	printf ("Contrast: M=toggle 'maximize contrast' mode\n");
	printf ("Brightness: '>'=brighter, '<'=darker (delta=10%%)\n");
	printf ("Verstical Scaling: '+'=increase, '-'=decrease (delta=10%%)\n");
	printf ("Resolution: Pn=draw only every nth line, for faster preview.\n");
	printf ("Settings: V\n");
	printf ("Quit: q\n");
}


static double ATAN2 (double y, double x)
{
	return ( ((x==0&&y==0) ? 0 : atan2(y,x)) );
}


