/* Terraform - (C) 1997-1998 Robert Gasch (Robert_Gasch@peoplesoft.com)
 *  - www.peoplesoft.com/peoplepages/g/robert_gasch/terraform/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#include "HeightFieldIO.h"
#include "GlobalTrace.h"
#include "GlobalSanityCheck.h"



/*
 *  constructor: initialize
 */
HeightFieldIO::HeightFieldIO (const char *fname) : FileIO (fname)
{
	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, "+++ HeightFieldIO\n");
}


/*
 *  destructor: initialize
 */
HeightFieldIO::~HeightFieldIO ()
{
	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, "--- HeightFieldIO\n");
}


/*
 *  getFileType: return the matching file type for a filename extension
 */
int HeightFieldIO::getFileType ()
{
	char 	*ext;

	if (!(ext=this->getExtension(false)))
		{
		printf ("Unable to determine extension of file [%s]\n", this->getName());
		return (-1);
		}
	else
	if (GlobalTrace::isSet (GlobalTrace::TRACE_FLOW))
		printf ("filename extension = %s\n", ext);

	if (!strcmp(ext,"gif")) 
		return (GIF);
	else if (!strcmp (ext,"pot")) 
		return (POT);
	else if (!strcmp (ext,"tga")) 
		return (TGA);
	else if (!strcmp (ext,"pgm")) 
		return (PGM);
	else if (!strcmp (ext,"pg8")) 
		return (PG8);
	else if (!strcmp (ext,"mat")) 
		return (MAT);
	else if (!strcmp (ext,"oct")) 
		return (OCT);
#ifdef HAVE_IMLIB
	else if (!strcmp (ext,"png")) 
		return (PNG);
	else if (!strcmp (ext,"tiff") || !strcmp (ext,"tif")) 
		return (TIFF);
#endif
	else if (!strcmp (ext,"dem")) 
		return (DEM);
	else if (!strcmp (ext,"gtop")) 
		return (GTP);
	else
		{ 
     		cerr << "Warning: " << ((ext) ? "unknown " : "missing ") << 
			"file extension (" << ext << ") ... defaulting to TGA \n";
		return (TGA);		// return default 
   		}
}


/*
 *  read: wrapper to call the proper read function 
 */
int HeightFieldIO::read (HeightField *HF)
{
	int			itype, rc=-1;
	HeightFieldReader 	*HFR;

	SanityCheck::bailout ((!HF), "HF==NULL", "HeightFieldIO::read");

	if (!this->exists())
		{
		printf ("HeightFieldIO::read(): file [%s] doesn't exist ...\n", this->getName());
		return (-1);
		}

	itype = this->getFileType ();
	if (itype == -1)
		return (itype);

	HFR = new HeightFieldReader (HF, this);

	if (itype == PGM || itype == PG8)
   		rc = HFR->readPGM ();  
	else
	if (itype == TGA)
		rc = HFR->readTGA ();
	else
	if (itype == MAT)
		rc = HFR->readMAT ();
	else
	if (itype == OCT)
		rc = HFR->readOCT ();
	else
	if (itype == GTP)
		rc = HFR->readGTOPO ();
	else
	if (itype == DEM)
		rc = HFR->readDEM ();
#ifdef HAVE_IMLIB
	else
	if (itype == PNG || itype == TIFF)
		rc = HFR->readImlib ();
#endif
	else
		printf ("HeightFieldIO::read(): file type %d not implemented ... cancelling read", itype);

	if (rc)
		printf ("Warning: File read [%s] returned an Error (%d)\n", this->getName(), rc);
	else
		HF->setLoadedFromFile (TRUE);

	delete HFR;
	return (rc);
} 


/*
 *  write: wrapper to call the proper write function 
 */
int HeightFieldIO::write (HeightField *HF)
{
	int			itype, rc=-1;
	HeightFieldWriter	*HFW;

	SanityCheck::bailout ((!HF), "HF==NULL", "HeightFieldIO::write");

	itype = this->getFileType ();
	if (itype == -1)
		return (itype);

	HFW = new HeightFieldWriter (HF, this);

	if (itype == TGA)
		rc = HFW->writeTGA ();
	else
	if (itype == PGM)
		rc = HFW->writePGM ();
	else
	if (itype == PG8)
		rc = HFW->writePG8 ();
	else
	if (itype == MAT)
		rc = HFW->writeMAT ();
	else
	if (itype == OCT)
		rc = HFW->writeOCT ();
	else
		printf ("HeightFieldIO::write(): file type %d not implemented ... cancelling write", itype);

	if (rc)
		printf ("Warning: File write [%s] returned an Error (%d)\n", this->getName(), rc);

	delete HFW;
	return (rc);
}

