/* Terraform - (C) 1997-1998 Robert Gasch (Robert_Gasch@peoplesoft.com)
 *  - www.peoplesoft.com/peoplepages/g/robert_gasch/terraform/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "HeightFieldIO.h"
#include "HeightFieldOps.h"
#include "GlobalTrace.h"
#include "GlobalSanityCheck.h"
#include "MathFunctions.h"
#include "MathGauss.h"
#include "MathTrig.h"
#include "Timer.h"


#define NO_FLOW		-987654321		// marker for 'undefined'
#ifndef MIN
# define MIN(a,b)	(a<b ? a : b)
#endif
#ifndef MAX
# define MAX(a,b)	(a>b ? a : b)
#endif 


/*
 *  constructor: initialize
 */
HeightFieldOps::HeightFieldOps (HeightField *HF) 
{
	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, "+++ HeightFieldOps\n");

	SanityCheck::bailout ((!HF), "HF==NULL", "HeightFieldOps::HeightFieldOps");
	SanityCheck::bailout ((!HF->d_hf), "HF->d_hf==NULL", "HeightFieldOps::HeightFieldOps");

	d_HF = HF;
}


/*
 *  destructor: nothing to clean up
 */
HeightFieldOps::~HeightFieldOps ()
{
	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, "--- HeightFieldOps\n");
}


/*
 *  upslopeSFD: find the number of single-flow-direction 
 *	upslope cells for a given cell. Cache is used to 
 * 	to store computed SFD values.  
 */
int HeightFieldOps::upslopeSFD (int x, int y, PTYPE *cache)
{
	int	ux=x, uy=y, 		// upflow x & y coordinates
		found, count=0,
		iLow, iHigh, 
		jLow, jHigh;

	SanityCheck::bailout ((!cache), "cache==NULL", "HeightFieldOps::upslopeSFD");

	if (cache[d_HF->Pos2Off(x,y)] == NO_FLOW)
		{
		found=FALSE;
		iLow = (ux > 0 ? ux-1 : 0);
		iHigh = (ux < (d_HF->d_xsize-1) ? ux+1 : d_HF->d_xsize-1);
		jLow = (uy > 0 ? uy-1 : 0);
		jHigh = (uy < (d_HF->d_ysize-1) ? uy+1 : d_HF->d_ysize-1);
		// find the highest downflow neighbour
		for (int i=iLow; i<=iHigh; i++)
		    for (int j=jLow; j<=jHigh; j++)
			{
			if (i != x || j!= y)
				{
				if (d_HF->d_hf[d_HF->Pos2Off(i,j)] > 
				    d_HF->d_hf[d_HF->Pos2Off(ux,uy)])
					{
					found=TRUE;
					ux=i;
					uy=j;
					}
				}
			}

		// downflow
		if (found)
		    count += upslopeSFD (ux,uy,cache)+1;
		else
		    count = 0;

		cache[d_HF->Pos2Off(x,y)] = count;
		d_cMiss++;
		}
	else
		{
		count = (int)cache[d_HF->Pos2Off(x,y)];
		d_cHit++;
		}

	return (count);
}


/*
 *  upslopeMFD: find the number of multiple-flow-direction
 *	upslope cells for a given cell. countedFlag should be 
 * 	re-initialized every time flowmap() calls this 
 *	function since it's being used as an indicator 
 * 	that a cell has been counted in the current MFD 
 *	pass for this cell. Since the 
 */
float HeightFieldOps::upslopeMFD (int x, int y, PTYPE *countedFlag)
{
	int	iLow, iHigh, 
		jLow, jHigh;
	float	count=0;

	SanityCheck::bailout ((!countedFlag), "countedFlag==NULL", "HeightFieldOps::upslopeMFD");

	iLow = (x > 0 ? x-1 : 0);
	iHigh = (x < (d_HF->d_xsize-1) ? x+1 : d_HF->d_xsize-1);
	jLow = (y > 0 ? y-1 : 0);
	jHigh = (y < (d_HF->d_ysize-1) ? y+1 : d_HF->d_ysize-1);
	for (int i=iLow; i<=iHigh; i++)
	    for (int j=jLow; j<=jHigh; j++)
		{
		if (i != x || j!= y)
			{
			if ( (d_HF->d_hf[d_HF->Pos2Off(i,j)] > 
			      d_HF->d_hf[d_HF->Pos2Off(x,y)])  &&
			     (countedFlag[d_HF->Pos2Off(i,j)] == NO_FLOW) )
				{
				countedFlag[d_HF->Pos2Off(i,j)] = TRUE;
				count += upslopeMFD (i,j,countedFlag) +
					 ((i==x || j==y) ? 0.6 : 0.4);
				}
			}
		}

	return (count);
}


/*
 *  flowpap: calculate the flowmap for the height field as detailed in 
 *    "Comparison of single and mulitple flow direction algorithms for 
 *    computing topographic parameters in TOPMODEL", Wolock & McCabe Jr., 
 *    Water Resources Research, Vol. 31, No. 5, Pages 1315-1324, May 1995
 *  The flowmap is essentially   ln(a/tan(b))   where 
 *	a = A/C 
 *	C = contour length
 * 	A = (number of upslope cells + 1) * (grid cell area)
 * 	tab(b) = ( change in elevation with neighboring cell / 
 * 		   cell center distance )
 *	ln = Napierian Log = 
 *			 	   log (n/10^7)
 *		ln (n) =     -	-------------------
 *				log (10^7/(10^7-1))
 * 
 *  Interestingly enough, the paper uses the Napierian Log, which 
 *  in my case gives huge numbers at the top of the scale and zero 
 *  at the bottom. I find that using the natural log I get a 
 *  range which closely resembles the ones shown in the paper. 
 *  Does anybody out there (who has found this) know more about this?
 *  I've probably goofed up somewhere but don't know where. 
 */
HeightField *HeightFieldOps::flowmap (bool doSFD, bool ignoreSealevel,
				float maxElvErode) 	// above this no erosion
{
	char	buf[256];
	PTYPE	ELV_RT = 0.33;
	int	i, j,
		offij,
		dfx, dfy, 			// upflow x & y coordinates
		found,	
		iLow, iHigh, 
		jLow, jHigh,
		dotint;				// dot intervals 
	double 	C=10,				// contour length XXXXXXX
		gca=C*C, 			// Grid Cell Area
		a, A,
		tanb;
	Timer	fTimer;
	float	eTime;				// elapsed time
	PTYPE	fmin=-NO_FLOW, fmax=NO_FLOW;	// min and max flowmap values
	PTYPE	*cache=NULL,			// SFD/MFD cache
		*hfFlow=NULL,			// flowmap 
		*initCache=NULL;		// cache used for memcpy source
	HeightField *HFF=NULL;			// resulting flowmap HF

	fTimer.start ();
	dotint = d_HF->d_size/50;
	d_cHit=d_cMiss=0;

	sprintf (buf, "Generating %s flowmap (%s) for %s (1 dot = %d cells) ...\n", 
		  (doSFD ? "SFD" : "MFD"), (ignoreSealevel ? "full" : "partial"), 
		  d_HF->d_name, dotint); 
	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, buf);

	hfFlow = new PTYPE[d_HF->d_size];	// resulting flowmap 
	cache = new PTYPE[d_HF->d_size];	// chache

	if (!doSFD)
		initCache = new PTYPE[d_HF->d_size];	// init values

	for (i=0; i<d_HF->d_size; i++)
		if (doSFD)
			cache[i] = NO_FLOW;
		else
			initCache[i] = NO_FLOW;
			
	if (!doSFD)
		memcpy (cache, initCache, sizeof(PTYPE)*d_HF->d_size);

	for (j=0; j<d_HF->d_ysize; j++)
		for (i=0; i<d_HF->d_xsize; i++)
		    if (ignoreSealevel ||
			(!ignoreSealevel && 
			  (d_HF->El(i,j) > d_HF->d_sealevel) &&
			  (d_HF->El(i,j) < maxElvErode) ) ) 
			{
			//printf ("%d, %d = ", i, j); fflush (stdout);

			if (doSFD)
				A = (upslopeSFD(i,j,cache)) * gca;
			else
				{
				memcpy (cache, initCache, 
					sizeof(PTYPE)*d_HF->d_size);
				A = (upslopeMFD(i,j,cache)) * gca;
				}
			a = A/C;

			// find lowest downslope square 
			found=FALSE;	
			dfx=i;
			dfy=j;
			iLow = (dfx > 0 ? dfx-1 : 0);
			iHigh = (dfx < (d_HF->d_xsize-1) ? dfx+1 : d_HF->d_xsize-1);
			jLow = (dfy > 0 ? dfy-1 : 0);
			jHigh = (dfy < (d_HF->d_ysize-1) ? dfy+1 : d_HF->d_ysize-1);
			for (int ii=iLow; ii<=iHigh; ii++)
			    for (int jj=jLow; jj<=jHigh; jj++)
				{
				if (ii != i || jj != j)
				    {
				    if ( d_HF->El(ii,jj) < d_HF->El(dfx,dfy) )
						{
						found = TRUE;
						dfx = ii;
						dfy = jj;
						}
				    }
				}

			// there is a downflow 
			offij = d_HF->Pos2Off(i,j);
			if (found && a)
				{
				tanb = (((d_HF->El(offij) - d_HF->El(dfx,dfy))*
					     (d_HF->d_xsize*ELV_RT)) / C);

				hfFlow[offij] = 
					//MathFunctions::napierianLog (a);
					//MathFunctions::napierianLog (1/tanb);
					//MathFunctions::napierianLog (a/tanb);
					log (a/tanb);	
					//log (a);	
				}
			// no downflow 
			else
				hfFlow[offij] = NO_FLOW;

			//printf ("Flow %d,%d, A=%f, flow=%f\n", i,j,A, hfFlow[offij]);
			if (hfFlow[offij] > fmax && hfFlow[offij]!=NO_FLOW)
				fmax = hfFlow[offij];
			if (hfFlow[offij] < fmin && hfFlow[offij]!=NO_FLOW)
				fmin = hfFlow[offij];

			if (!(offij%dotint) &&
			    GlobalTrace::trace(GlobalTrace::TRACE_VERBOSE, "."))
				;
			}

	// make 0 lowest value so that NO_FLOW==0 
	for (int i=0; i<d_HF->d_size; i++)
		{
		if (hfFlow[i] == NO_FLOW)
			hfFlow[i] = 0; 
		else
			hfFlow[i] -= fmin; 
		}
			

	fTimer.stop ();
	eTime = fTimer.getElapsedSeconds ();
	sprintf (buf, "\nDone. Elapesed time: %.2f seconds\n", eTime);
	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, buf);

	sprintf (buf, "Flowmap max, min = %f, %f\nCache hit, miss = %ld, %ld\n", 
		fmax, fmin, d_cHit, d_cMiss);
	GlobalTrace::trace (GlobalTrace::TRACE_DEBUG, buf);

	delete [] cache;
	if (!doSFD)
		delete [] initCache;

	HFF = new HeightField (hfFlow, d_HF->d_xsize, d_HF->d_ysize);
	HFF->gatherStatistics ();
	HFF->setSaved (FALSE);

	//HeightFieldIO           *HFIO;
	//HFIO = new HeightFieldIO (buf);
	//HFIO->write (HFF);
	//delete HFIO;

	return HFF;
}


/* 
 * erode: erode the HF using the flowmap to determine the terrain delta
 */
int HeightFieldOps::erode (int 	nTurns, 	// # of erodsion turns
			   int 	maxFlowAge, 	// new flowmap every n turns
			   float maxElvErode, 	// above this no erosion
			   char *fName,		// if fName = save every maxFlowAge turns
			   bool trimLocalPeaks,	// trim if (flowmap==0)
			   bool doSFD, 		// flowmap param
			   bool ignoreSealevel, // flowmap param
			   HeightFieldDraw *HFD)
{
	char		buf[256];
	HeightField 	*HFF;			// flowmap HF
	int		nFlowmap=0,		// number of flowmaps so far
			flowAge=0,		// age of flowmap
			dotint;
	double		errate = 2e-4;		// max erosion per turn

	dotint = int(maxFlowAge/50.0);

	SanityCheck::bailout ((!HFD), "HFD==NULL", "HeightFieldOps::erode");
	// fName should be formatted as a sprintf format string and contain %d
	if (fName)
		{
		SanityCheck::bailout ((!strstr(fName,"%d")), "bad string format", "HeightFieldOps::erode");
		SanityCheck::bailout ((!strchr(fName, '.')), "bad string format", "HeightFieldOps::erode");
		}

	HFF = flowmap (doSFD, ignoreSealevel);
	for (int i=0; i<nTurns; i++, flowAge++)
		{
		if (!(i%dotint) &&
		    GlobalTrace::trace(GlobalTrace::TRACE_FLOW, "."))
			;

		d_HF->d_hf[0] -= errate*HFF->d_hf[0];
		for (int n=1; n<d_HF->d_size; n++)
			{
			if (HFF->d_hf[n] > 0)
				d_HF->d_hf[n] -= errate*HFF->d_hf[n];
			else
				{
				// average flowmap vale with neighbours
				if (trimLocalPeaks)
					d_HF->d_hf[n] -= errate * 
						( HFF->El(n-1) + 
						  HFF->El(n+1) + 
						  HFF->El(n-d_HF->d_xsize) + 
						  HFF->El(n+d_HF->d_xsize) )/4;
				else
					d_HF->d_hf[n] -= errate*HFF->El(n);
				}

			if (d_HF->d_hf[n] < 0)
				d_HF->d_hf[n] = 0;
			}

		if ((maxFlowAge>0) && (flowAge == maxFlowAge))
			{
			GlobalTrace::trace (GlobalTrace::TRACE_FLOW, "\n");
			//roughSmooth (false, true, 1.0);
			d_HF->gatherStatistics ();
			if (HFD)
				HFD->draw ();

			if (fName)
				{
				sprintf (buf, fName, nFlowmap);
				HeightFieldIO           *HFIO;

				HFIO = new HeightFieldIO (buf);
				HFIO->write (d_HF);
				delete HFIO;  
				}

			delete HFF;
			HFF = flowmap (doSFD, ignoreSealevel);
			flowAge = 0;
			nFlowmap++;

			sprintf (buf, "\nEroding %d/%d\n", i, nTurns); 
			GlobalTrace::trace (GlobalTrace::TRACE_FLOW, buf);
			}
		}

	if (HFF)
		delete HFF;

	return 0;
}



/*
 * hfDouble()  double resolution using modified-midpoint-displacement
 *	       algorithm; different versions for tiling and non-tiling
 *	       matrices. <f> is local fractal scaling, <f2> is overall
 *	       fractal scaling.
 * 	       This routine is adapted from John Beals's HF-Lab.
 */
int HeightFieldOps::hfDouble (double f, double f2) 	
{
	char		buf[80];
	int 		x, y;
	int 		xsize=0, ysize=0;
	int 		xsize1, ysize1;   	// old1 HF 
	int 		xsize2, ysize2;   	// old2 HF
  	PTYPE 		*hf1=d_HF->d_hf, 
			*hf2;
  	int 		tile=FALSE;
  	D 		sfac;
  	D 		a, b, c, d;
	MathGauss	*mGauss = new MathGauss (time(NULL));

	xsize1 = d_HF->d_xsize;
	ysize1 = d_HF->d_ysize;

	if (!tile) 			// non-tiling: don't wrap at edges 
		{  
		xsize2--;
		ysize2--;
		}

	d_HF->d_xsize = xsize2 = xsize1*2;
	d_HF->d_ysize = ysize2 = ysize1*2;
	d_HF->d_size = xsize2*ysize2;

	sprintf (buf, "doubling to %dx%d ... ", xsize2, ysize2);  
	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, buf);

	hf2 = new PTYPE[d_HF->d_size];
	d_HF->d_hf = hf2;

	f2 /= sqrt((D)d_HF->d_size);  /* normalize to matrix size */

	for (y=0;y<ysize1;y++) 
     		for (x=0;x<xsize1;x++)
      			hf2[d_HF->Pos2Off(x*2,y*2)] = hf1[xsize1*y+x];

	if (!tile) 			// fill in edges 
		{
		for (y=1;y<ysize2;y+=2) 
			{
			hf2[d_HF->Pos2Off(0,y)] = hf2[d_HF->Pos2Off(0,y-1)];
			hf2[d_HF->Pos2Off(xsize2-1,y)] = hf2[d_HF->Pos2Off(xsize2-1,y-1)];
			}
		for (x=1;x<xsize2;x+=2) 
			{
			hf2[d_HF->Pos2Off(x,0)] = hf2[d_HF->Pos2Off(x-1,0)];
			hf2[d_HF->Pos2Off(x,ysize2-1)] = hf2[d_HF->Pos2Off(x-1,ysize2-1)];
			}
		}

	xsize = xsize2;
	ysize = ysize2;
	for (y=1; y<ysize2; y+=2) 	// tiling and non-tiling interpolation
	    {
	    for (x=1; x<xsize2; x+=2) 
		{
		if (tile) 
			{
			a = d_HF->Elmod(x+1,y+1); b = d_HF->Elmod(x-1,y+1);
			c = d_HF->Elmod(x+1,y-1); d = d_HF->Elmod(x-1,y-1);
			} 
		else 
			{
			a = d_HF->Elclip(x+1,y+1); b = d_HF->Elclip(x-1,y+1);
			c = d_HF->Elclip(x+1,y-1); d = d_HF->Elclip(x-1,y-1);
			}
	        sfac = f2+f*(MAX(MAX(a,b),MAX(c,d)) - MIN(MIN(a,b),MIN(c,d)));
		d_HF->d_hf[d_HF->Pos2Off(x,y)] = (a+b+c+d)/4.0+sfac*(mGauss->rnd()-0.5);
		}
	    }

	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, "cent1 ...");

	for (y=0; y<ysize2; y+=2) 
	    {      
	    for (x=1; x<xsize2; x+=2) 
		{
		if (tile) 
			{
			a = d_HF->Elmod(x+1,y); b = d_HF->Elmod(x-1,y);
			c = d_HF->Elmod(x,y+1); d = d_HF->Elmod(x,y-1);
			} 
		else 
			{
			a = d_HF->Elclip(x+1,y); b = d_HF->Elclip(x-1,y);
			c = d_HF->Elclip(x,y+1); d = d_HF->Elclip(x,y-1);
			}
		sfac = f2+f*(MAX(MAX(a,b),MAX(c,d)) - MIN(MIN(a,b),MIN(c,d)));
		d_HF->d_hf[d_HF->Pos2Off(x,y)] = (a+b+c+d)/4.0+sfac*(mGauss->rnd()-0.5);
		}
	    }

	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, "edge1 ...");

	for (y=1; y<ysize2; y+=2) 
	    {      
	    for (x=0; x<xsize2; x+=2) 
		{
		if (tile) 
			{
			a = d_HF->Elmod(x+1,y); b = d_HF->Elmod(x-1,y);
			c = d_HF->Elmod(x,y+1); d = d_HF->Elmod(x,y-1);
			} 
		else 
			{
			a = d_HF->Elclip(x+1,y); b = d_HF->Elclip(x-1,y);
			c = d_HF->Elclip(x,y+1); d = d_HF->Elclip(x,y-1);
			}
		sfac = f2+f*(MAX(MAX(a,b),MAX(c,d)) - MIN(MIN(a,b),MIN(c,d)));
		d_HF->d_hf[d_HF->Pos2Off(x,y)] = (a+b+c+d)/4.0+sfac*(mGauss->rnd()-0.5);
		}
	    }

	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, "edge2 ...");

	for (y=0;y< ysize2; y+=2) 
	    {      /* backtrack and adjust original data */
	    for (x=0; x<xsize2; x+=2) 
		{
		if (tile) 
			{
			a = d_HF->Elmod(x+1,y+1); b = d_HF->Elmod(x-1,y+1);
			c = d_HF->Elmod(x+1,y-1); d = d_HF->Elmod(x-1,y-1);
			} 
		else 
			{
			a = d_HF->Elclip(x+1,y+1); b = d_HF->Elclip(x-1,y+1);
			c = d_HF->Elclip(x+1,y-1); d = d_HF->Elclip(x-1,y-1);
			}
		sfac = 0.5*(f2+f*(MAX(MAX(a,b),MAX(c,d)) - MIN(MIN(a,b),MIN(c,d))));
		d_HF->d_hf[d_HF->Pos2Off(x,y)] = (a+b+c+d)/4.0+sfac*(mGauss->rnd()-0.5);
		}
	    }

	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, "done\n");

	delete [] hf1;
	delete mGauss;

	d_HF->gatherStatistics ();
	d_HF->setSaved (FALSE);
	return (0);
}  



/*
 * hfHalf ()   half resolution by averaging 4 points into 1
 * 	       This routine is adapted from John Beals's HF-Lab.
 */
int HeightFieldOps::hfHalf() 
{
	int 	x,y,xx,yy;
	int 	xsize1, ysize1;   /* old1 HF */
	int 	xsize2, ysize2;   /* old2 HF */
  	PTYPE 	*hf1=d_HF->d_hf, *hf2;
	char	buf[80];

#define P2O(x,y)	y*xsize1+x

	if (d_HF->d_xsize*d_HF->d_ysize < 25)
		return 1;

	xsize1 = d_HF->d_xsize;
	ysize1 = d_HF->d_ysize;
	d_HF->d_xsize = xsize2 = xsize1/2;
	d_HF->d_ysize = ysize2 = ysize1/2;
	d_HF->d_size = xsize2*ysize2;

	sprintf(buf, "Diminishing to %dx%d ... ", xsize2,ysize2); 
	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, buf);

	hf2 = new PTYPE[d_HF->d_size];
	d_HF->d_hf = hf2;

	for (y=0; y<ysize2; y++) 
	    {      /* copy over initial values */
	    yy = y * 2;
	    for (x=0; x<xsize2; x++) 
		{
		xx = x * 2;
		d_HF->d_hf[d_HF->Pos2Off(x,y)] = 
			( hf1[P2O(xx,yy)] + hf1[P2O(xx+1,yy)] +
			  hf1[P2O(xx,yy+1)] + hf1[P2O(xx+1,yy+1)] ) /4.0;
		}
	    }

	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, "done\n");

	delete [] hf1;
	d_HF->gatherStatistics ();
	d_HF->setSaved (FALSE);
	return (0);
}  


/*
 * hfResize()   half resolution by averaging 4 points into 1
 * 	       This routine is adapted from John Beals's HF-Lab.
 */
HeightField *HeightFieldOps::hfResize (int newx, int newy) 
{
	float 		incx = d_HF->d_xsize/newx,
			incy = d_HF->d_ysize/newy,
			offx, 
			offy;
	int		offSrc,
			offDst;
	HeightField	*HF=NULL;

	if (d_HF->d_xsize<=newx || d_HF->d_ysize<newy)
		return NULL;

	PTYPE	*hf = new PTYPE[newx*newy];

	for (offy=0; offy<newy; offx+=incy)
		for (offx=0; offx<newx; offx+=incx)
			{
			offSrc = ( (int)(offy*incy*newx) + (int)(offx*incx) );
			offDst = (int) (((int)offy)*newx+offx);
			hf[offDst] = d_HF->El (offSrc);
			}

	HF = new HeightField (hf, newx, newy, d_HF->getName());
	HF->setSaved (FALSE);

	return HF;
}



/*
 *  placePOVObject: place a PovRay object on the landscape
 * 	We use the following method:
 *	- calculate a basic square size dependant upon object placement density and HF scale size in POV units.
 *	- calculate the size of an inner square proportional to pvar 
 *	- place an object of randomized size at the correct offset for the inner square
 */
int HeightFieldOps::placePOVObject (
			char	*outfile,	// output filename
			char 	*object, 	// object name
			int	hfsx,		// height field x scaling factor
			int	hfsy,		// height field y scaling factor
			int	hfsz,		// height field y scaling factor
			PTYPE 	floor, 		// low elevation for object placement
			PTYPE	ceil, 		// high elevation for object placement
			float	density,	// density factor for object placement
			float	pvar,		// placement variance ratio 
			int	osx, 		// object x scaling factor
			int	osy, 		// object y scaling factor
			int	osz, 		// object y scaling factor
			PTYPE	osxvar, 	// object x scaling factor variance
			PTYPE	osyvar, 	// object z scaling factor variance
			PTYPE	oszvar, 	// object z scaling factor variance
			int	seed)		// random number seed
{
	FILE	*fout, *ft;
	unsigned int s=(unsigned int)seed;	// seed
	float	e,				// elevation
		xOff, yOff,			// offset for inner placement square
		nGridX, nGridY,			// numbers of x and y size of a grid
		gridX, gridY,			// x and y size of a grid
		ixpov, iypov, 			// x and y increment POVray scale
		ixhf, iyhf, 			// x and y increment d_HF subscript scale
		hsx, hsy, hsz, lsx, lsy, lsz;	// high and low scaling factors
	int	nobs=0;
	char	buf[80];

	//printf ("\n%s, %s, hfs[xyz]=%d, %d, %d\n", outfile, object, hfsx, hfsy, hfsz);
	//printf ("floor=%f, ceil=%f, dens=%f, pvar=%f,\n", floor, ceil, density, pvar);
	//printf ("os[xyz]=%d,%d,%d ", osx, osy, osz);
	//printf ("os[xyz]var=%f, %f, %f, seed=%d\n\n", osxvar, osyvar, oszvar, seed);

	// Sanity Checks
	SanityCheck::bailout ((!outfile), "outfile==NULL", "HeightFieldOps::PlacePOVObject");
	SanityCheck::bailout ((!strlen(outfile)), "strlen(outfile)==0", "HeightFieldOps::PlacePOVObject");
	SanityCheck::bailout ((!object), "object==NULL", "HeightFieldOps::PlacePOVObject");
	SanityCheck::bailout ((!strlen(object)), "strlen(object)==0", "HeightFieldOps::PlacePOVObject");
	SanityCheck::bailout ((floor < d_HF->d_min || floor > d_HF->d_max), 
			"floor out of range", "HeightFieldOps::PlacePOVObject");
	SanityCheck::bailout ((ceil < d_HF->d_min || ceil > d_HF->d_max), 
			"ceil out of range", "HeightFieldOps::PlacePOVObject");
	SanityCheck::bailout ((density <= 0), "density<=0", "HeightFieldOps::PlacePOVObject");
	SanityCheck::bailout ((pvar < 0 || pvar > 1), "pvar out of range", "HeightFieldOps::PlacePOVObject");
	SanityCheck::bailout ((osxvar < 0 || osxvar > 1), "osxvar out of range", "HeightFieldOps::PlacePOVObject");
	SanityCheck::bailout ((osyvar < 0 || osyvar > 1), "osyvar out of range", "HeightFieldOps::PlacePOVObject");
	SanityCheck::bailout ((oszvar < 0 || oszvar > 1), "oszvar out of range", "HeightFieldOps::PlacePOVObject");
	SanityCheck::bailout ((seed < 0), "seed<0", "HeightFieldOps::PlacePOVObject");

	if (!(fout=fopen (outfile, "w"))) 		// open output file
		{
		cerr << "HeightFieldOps::PlacePOVObject(): can't open file " << outfile << "\n";
		return (-1);
		}

	sprintf (buf, "Placing PovObjects on %s ... ", d_HF->d_name);
	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, buf);

	// init random number generator		
	if (s)
		srand (s);

	// find percentage scaling ranges (0 .. 2)
	hsx=1+osxvar; hsy=1+osyvar; hsz=1+oszvar;
	lsx=1-osxvar; lsy=1-osyvar; lsz=1-oszvar;

	// set number of cells
	nGridX=(hfsx/osx)*density;	// high density == more cells
	nGridY=(hfsz/osz)*density;	// povray dim z == program dim y

	// set cell size 
	gridX=(hfsx/nGridX);
	gridY=(hfsz/nGridY);

	// calculate max offset from center
	xOff=(gridX-(gridY-(gridX*pvar)));
	yOff=(gridY-(gridY-(gridY*pvar)));

	printf ("%f, %f, 	%f, %f\n", nGridX, nGridY, gridX, gridY); 
	printf ("%f, %f\n", xOff, yOff);

	ft=stdout;
	for (int c=0; c<2; c++)
	   {
	   if ((c==0 && GlobalTrace::isSet (GlobalTrace::TRACE_DEBUG)) || c==1)
		{
		if (c==1)
			fprintf (ft, "// File created by terraform ...\n");
		fprintf (ft, "// HeightFieldOps::PlacePOVObject Settings\n");
		fprintf (ft, "// Output Filename: 		%s\n", outfile);
		fprintf (ft, "// Object name:	 		%s\n", object);
		fprintf (ft, "// Height Field Scales (x,y,z):	%d, %d, %d\n", hfsx, hfsy, hfsz);
		fprintf (ft, "// Floor, Ceil, Density:	%f, %f, %f\n", floor, ceil, density);
		fprintf (ft, "// x scale & variance:		%d, %f\n", osx, osxvar);
		fprintf (ft, "// y scale & variance:		%d, %f\n", osy, osyvar);
		fprintf (ft, "// z scale & variance:		%d, %f\n", osz, oszvar);
		fprintf (ft, "// seed:			%d\n", seed);
		fprintf (ft, "// hsx, hsy, hsz:			%f, %f, %f\n", hsx, hsy, hsz);
		fprintf (ft, "// lsx, lsy, lsz:			%f, %f, %f\n", lsx, lsy, lsz);
		fprintf (ft, "// nGridX, nGridY:		%f, %f\n", nGridX, nGridY);
		fprintf (ft, "// xOff, yOff:			%f, %f\n\n", xOff, yOff);
		}
	   ft=fout;
	   }

	ixpov=gridX;				// increment on POVray scale
	iypov=gridY;				// povray dim z == array dim y
	ixhf=d_HF->d_xsize/nGridX;		// increment on d_HF subscript scale
	iyhf=d_HF->d_ysize/nGridY;		// increment on d_HF subscript scale

	
	printf ("xOff=%f, yOff=%f\nnGridX=%f, nGridY=%f\n", xOff, yOff, nGridX, nGridY);
	printf ("ixpov=%f, iypov=%f, ixhf=%f, iyhf=%f\n", ixpov, iypov, ixhf, iyhf);

	// loop through the squares  
	for (float jp=0, jh=0; jp<hfsy; jp+=iypov, jh+=iyhf)
		for (float ip=0, ih=0; ip<hfsx; ip+=ixpov, ih+=ixhf)
			{
			float	aoffx, aoffz,		// actual placement offset in grid 
				aosx=0, aosy=0, aosz=0,	// actual object size scale factor x, y, z
				aopx=0, aopy=0, aopz=0;	// actual object position x, y, z

			// check height at grid center
			//printf ("X:: %f, %f, %f, %f, %d", ip, jp, ih, jh, off);fflush(stdout);
			e=d_HF->d_hf[d_HF->Pos2Off(ip,jp)];
			if (e >= floor && e <= ceil)
				{
				nobs++;

				// get an offset for the inner placement square
				aoffx = xOff*rand()/(RAND_MAX+1.0);
				aoffz = yOff*rand()/(RAND_MAX+1.0);

				// calculate the object position in the scene
				if (rand() > 0.5)
					aopx = ip+ip/2+aoffx;
				else
					aopx = ip+ip/2-aoffx;

				if (rand() > 0.5)
					aopz = jp+jp/2+aoffz;
				else
					aopz = jp+jp/2-aoffz;

				aopy = e*hfsz;
	
				// calculate the object scaling factors
				aosx = ((hsx-lsx)*rand()/(RAND_MAX+1.0)+lsx);
				aosy = ((hsy-lsy)*rand()/(RAND_MAX+1.0)+lsy);
				aosz = ((hsz-lsz)*rand()/(RAND_MAX+1.0)+lsz);

				printf ("+\n");
				fprintf (fout, "object { %s  scale < %f, %f, %f > translate < %f, %f, %f > }\n",
					object, osx*aosx, osy*aosy, osz*aosz, aopx, aopy, aopz);
				}
			else
				printf ("-\n");
			}
	fclose (fout); 			// close output file

	sprintf (buf, "%d %ss placed\n", nobs, object);
	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, buf);

	return (0);
}


/*
 *  roughSmooth: either roughen or smooth the height field. 
 */
int HeightFieldOps::roughSmooth (bool roughen, bool bigGrid, float factor)
{
	SanityCheck::bailout ((factor<0), "factor<0", "HeightFieldOps::smooth");
	SanityCheck::bailout ((factor>1), "factor>1", "HeightFieldOps::smooth");

	PTYPE	o, t;
	int	limx=d_HF->d_xsize-1, limy=d_HF->d_ysize-1;

	for (int x=1; x<limx; x++)
		for (int y=1; y<limy; y++)
			{
			o = d_HF->El (x, y);
			if (bigGrid == 0)
				{
				t = d_HF->El(x-1,y) + d_HF->El(x+1,y) +
				    d_HF->El(x,y-1) + d_HF->El(x,y+1);
				t /= 4;
				}
			else
				{
				t = d_HF->El(x-1,y) + d_HF->El(x+1,y) +
				    d_HF->El(x,y-1) + d_HF->El(x,y+1) + 
				    d_HF->El(x-1,y-1) + d_HF->El(x-1,y+1) +
				    d_HF->El(x+1,y-1) + d_HF->El(x+1,y+1);
				t /= 8;
				}
			if (roughen)
				d_HF->d_hf[d_HF->Pos2Off(x,y)] = o-factor*(t-o);
			else
				d_HF->d_hf[d_HF->Pos2Off(x,y)] = o+factor*(t-o);
			}

	d_HF->gatherStatistics ();
	d_HF->setSaved (FALSE);
	return (0);
}


/*
 *  transform: transform the elevation by some power factor
 */
int HeightFieldOps::transform (float factor, int adjust_sea_level)
{
	char	buf[80];

	SanityCheck::bailout ((!d_HF->d_hf), "d_HF->d_hf==NULL", "HeightFieldOps::transform");

	sprintf (buf, "Transforming with power %f ...\n", factor);
	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, buf);

	for (int i=0; i<d_HF->d_size; i++)
		d_HF->d_hf[i]=pow (d_HF->d_hf[i], factor);
	if (adjust_sea_level)
		d_HF->d_sealevel=pow (d_HF->d_sealevel, factor);
	d_HF->gatherStatistics ();
	d_HF->setSaved (FALSE);
	return (0);
}


/*
 *  rotate90: counterclockwise rotate in 90, 180 or 270 degrees
 */
int HeightFieldOps::rotate90 (int degrees)
{
	PTYPE	*hft=NULL;
	char	buf[80];

	//SanityCheck::bailout ((!hft), "hft==NULL", "HeightFieldOps::Rotate90");
	SanityCheck::bailout ((!d_HF->d_hf), "d_HF->d_hf==NULL", "HeightFieldOps::rotate90");

	sprintf (buf, "Rotating %d degrees ...\n", degrees); 
	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, buf);

	if (degrees < 0)
		degrees = (-degrees)%360;

	if (degrees > 360)
		degrees = degrees%360;

	if (degrees == 90)
		{
		hft=new PTYPE [d_HF->d_size];
		for (int x=0; x<d_HF->d_xsize; x++)
			for (int y=0; y<d_HF->d_ysize; y++)
				hft[d_HF->Pos2Off (x,y)]=d_HF->d_hf[d_HF->Pos2Off (d_HF->d_ysize-1-y, x)];

		delete [] d_HF->d_hf;
		d_HF->d_hf=hft;
		}
	else
	if (degrees == 180)
		{
		PTYPE	pt;
		int	idx=d_HF->d_size-1, 
			lim=d_HF->d_size/2, 
			c=0;

		while (c < lim)
			{
			pt=d_HF->d_hf[c];
			d_HF->d_hf[c]=d_HF->d_hf[idx-c];
			d_HF->d_hf[idx-c]=pt;
			c++;
			}
		}
	else
	if (degrees == 270)
		{
		hft=new PTYPE [d_HF->d_size];
		for (int i=0; i<3; i++)
			{
			for (int x=0; x<d_HF->d_xsize; x++)
				for (int y=0; y<d_HF->d_ysize; y++)
					hft[d_HF->Pos2Off (x,y)]=d_HF->d_hf[d_HF->Pos2Off (d_HF->d_ysize-1-y, x)];
			if (i<2)
				memcpy (d_HF->d_hf, hft, d_HF->d_size*sizeof(PTYPE));
			}
		delete [] d_HF->d_hf;
		d_HF->d_hf=hft;
		}
	else
		{
		printf ("Warning: Rotate90() paramter (%d) out of range ... Cancelling \n", degrees);
		return (-1);
		}

	d_HF->setSaved (FALSE);
	return (0);
}


/*
 *  Invert: invert the height field 
 */
int HeightFieldOps::invert (int adjust_sea_level)
{
	char 	buf[80];
	SanityCheck::bailout ((!d_HF->d_hf), "d_HF->d_hf==NULL", "HeightFieldOps::invert");

	sprintf (buf, "Inverting height field ... \n");
	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, buf);

	for (int i=0; i<d_HF->d_size; i++)
		d_HF->d_hf[i]=1-d_HF->d_hf[i];
	if (adjust_sea_level)
		d_HF->d_sealevel=1-d_HF->d_sealevel;
	d_HF->gatherStatistics ();
	d_HF->setSaved (FALSE);
	return (0);
}


/*
 *  hf_sort: compare 2 PTYPE values; used by qsort()
 */
inline int HeightFieldOps::hf_sort (const void *_v1_, const void *_v2_)
{
	PTYPE	*v1 = static_cast<PTYPE*>(_v1_),
		*v2 = static_cast<PTYPE*>(_v2_);

	if (*v1 < *v2)
		return (-1);
	else
	if (*v1 > *v2)
		return (1);
	else
		return (0);
}


/*
 *  fold: fold height field edges under sealevel.
 *  How: Take a slice 'margin' long, fill it with the ABS 
 *  difference between it's height and the HF, sort it and 
 *  then subtract it from the HF elevation at that point 
 *  (starting with the largest values on the HF outside), 
 *  scaling the value subtracted in a linear fashion (0..1).
 */
int HeightFieldOps::fold (int margin)
{
	int	i=0, lim=0, 
		(*hf_cmp_func) (const void *, const void *) = hf_sort;
	float	pstep; 
	PTYPE	slice[margin]; 
	char 	buf[80];

	SanityCheck::bailout ((!d_HF->d_hf), "d_HF->d_hf==NULL", "HeightFieldOps::Fold");
	SanityCheck::bailout ((!margin), "margin==0", "HeightFieldOps::Fold");
	SanityCheck::bailout ((margin>d_HF->d_xsize/2 || margin>d_HF->d_ysize/2), "margin>x/2 || margin>y/2", "HeightFieldOps::Fold");

	sprintf (buf, "Folding with margin %d ...\n", margin); 
	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, buf);

	// top and bottom 
	for (int x=0; x<d_HF->d_xsize; x++)
	    {
	    if (d_HF->d_hf[d_HF->Pos2Off(x,0)] > d_HF->d_sealevel)
		{
		for (lim=0; lim<margin && d_HF->d_hf[d_HF->Pos2Off(x,lim)]>d_HF->d_sealevel; lim++)
			slice[lim] = d_HF->d_hf[d_HF->Pos2Off(x,lim)] - d_HF->d_sealevel;
		qsort ((void *)slice, (size_t)lim, (size_t)sizeof(PTYPE), hf_cmp_func);
		pstep = (float)1/(float)lim;
		for (i=0; i<lim; i++)
			{
			d_HF->d_hf[d_HF->Pos2Off(x,i)] -= (slice[lim-1-i]*pstep*(lim-i));
			if (d_HF->d_hf[d_HF->Pos2Off(x,i)] < 0)
				d_HF->d_hf[d_HF->Pos2Off(x,i)] = 0;
			}
				
		}

	    if (d_HF->d_hf[d_HF->Pos2Off(x,d_HF->d_ysize-1)] > d_HF->d_sealevel)
		{
		for (lim=0; lim<margin && d_HF->d_hf[d_HF->Pos2Off(x,d_HF->d_ysize-1-lim)]>d_HF->d_sealevel; lim++)
			slice[lim] = d_HF->d_hf[d_HF->Pos2Off(x,d_HF->d_ysize-1-lim)] - d_HF->d_sealevel;
		qsort ((void *)slice, (size_t)lim, (size_t)sizeof(PTYPE), hf_cmp_func);
		pstep = (float)1/(float)lim;
		for (i=0; i<lim; i++)
			{
			d_HF->d_hf[d_HF->Pos2Off(x,d_HF->d_ysize-1-i)] -= (slice[lim-1-i]*pstep*(lim-i));
			if (d_HF->d_hf[d_HF->Pos2Off(x,d_HF->d_ysize-1-i)] < 0)
				d_HF->d_hf[d_HF->Pos2Off(x,d_HF->d_ysize-1-i)] = 0;
			}
		}
	    }

	// left and right sides 
	for (int y=0; y<d_HF->d_ysize; y++)
	    {
	    if (d_HF->d_hf[d_HF->Pos2Off(0,y)] > d_HF->d_sealevel)
		{
		for (lim=0; lim<margin && d_HF->d_hf[d_HF->Pos2Off(lim,y)]>d_HF->d_sealevel; lim++)
			slice[lim] = d_HF->d_hf[d_HF->Pos2Off(lim,y)] - d_HF->d_sealevel;
		qsort ((void *)slice, (size_t)lim, (size_t)sizeof(PTYPE), hf_cmp_func);
		pstep = (float)1/(float)lim;
		for (i=0; i<lim; i++)
			{
			d_HF->d_hf[d_HF->Pos2Off(i,y)] -= (slice[lim-1-i]*pstep*(lim-i));
			if (d_HF->d_hf[d_HF->Pos2Off(i,y)] < 0)
				d_HF->d_hf[d_HF->Pos2Off(i,y)] = 0;
			}
		}

	    if (d_HF->d_hf[d_HF->Pos2Off(d_HF->d_xsize-1,y)] > d_HF->d_sealevel)
		{
		for (lim=0; lim<margin && d_HF->d_hf[d_HF->Pos2Off(d_HF->d_xsize-1-lim,y)]>d_HF->d_sealevel; lim++)
			slice[lim] = d_HF->d_hf[d_HF->Pos2Off(d_HF->d_xsize-1-lim,y)] - d_HF->d_sealevel;
		qsort ((void *)slice, (size_t)lim, (size_t)sizeof(PTYPE), hf_cmp_func);
		pstep = (float)1/(float)lim;
		for (i=0; i<lim; i++)
			{
			d_HF->d_hf[d_HF->Pos2Off(d_HF->d_ysize-1-i,y)] -= (slice[lim-1-i]*pstep*(lim-i));
			if (d_HF->d_hf[d_HF->Pos2Off(d_HF->d_ysize-1-i,y)] < 0)
				d_HF->d_hf[d_HF->Pos2Off(d_HF->d_ysize-1-i,y)] = 0;
			}
		}
	    }

	d_HF->gatherStatistics ();
	d_HF->setSaved (FALSE);
	return (0);
}


/*
 *  linearScale: pick a point (centerx, centery) and scale the points 
 *  where distance is mindist<=distance<=maxdist linarly from either 
 *  1..0 if scale_factor is < 1; otherwise was scale from 1..2. 
 */
int HeightFieldOps::linearScale (	float 	centerx, 
					float	centery, 
					float 	scale_factor, 
					float 	mindist,
					float	maxdist, 
					float 	smoothFactor,
					bool 	invertScale,
					bool	scaleFullDist)
{
	float	p,		// percentage change multiplication factor
		smoothFloor,
		smoothCeil;
	double  d, 		// distance
		md;		// max distance
	char	buf[80];
		

	SanityCheck::bailout ((!d_HF->d_hf), "d_HF->d_hf==NULL", "HeightFieldOps::LinearScale");
	SanityCheck::bailout ((centerx<0), "centerx < 0", "HeightFieldOps::LinearScale");
	SanityCheck::bailout ((centerx>1), "centerx > 1", "HeightFieldOps::LinearScale");
	SanityCheck::bailout ((centery<0), "centery < 0", "HeightFieldOps::LinearScale");
	SanityCheck::bailout ((centery>1), "centery > 1", "HeightFieldOps::LinearScale");
	SanityCheck::bailout ((scale_factor<0), "scale_factor < 0", "HeightFieldOps::LinearScale");
	SanityCheck::bailout ((mindist<0), "mindist < 0", "HeightFieldOps::LinearScale");
	SanityCheck::bailout ((mindist>1), "mindist > 1", "HeightFieldOps::LinearScale");
	SanityCheck::bailout ((maxdist<0), "maxdist < 0", "HeightFieldOps::LinearScale");
	SanityCheck::bailout ((maxdist>1), "maxdist > 1", "HeightFieldOps::LinearScale");
	SanityCheck::bailout ((smoothFactor<0), "smoothFactor < 0", "HeightFieldOps::LinearScale");
	SanityCheck::bailout ((smoothFactor>1), "smoothFactor > 1", "HeightFieldOps::LinearScale");

	if (SanityCheck::warning ((maxdist==mindist), "maxdist==mindist - linear scale cancelled", NULL))
		return 1;

	sprintf (buf, "Linear Scaling (%.2f, %.2f, %.2f, %.2f, %.2f, %s, %s) ...\n",
		centerx, centery, scale_factor, mindist, maxdist, 
		invertScale==TRUE ? "InvertScale" : "NoInvertScale",
		scaleFullDist?"ScaleFullDist":"NoScaleFullDist");
	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, buf);

	// convert to real (subscript) scale
	centerx *= d_HF->d_xsize;
	centery *= d_HF->d_ysize;
	mindist *= (centerx > centery ? centerx : centery);
	maxdist *= (centerx > centery ? centerx : centery);
	smoothFloor = mindist*(1-smoothFactor);
	smoothCeil = maxdist*(1+smoothFactor);

	md=MathTrig::distance (centerx, centery, 0, 0);
	if ((d=MathTrig::distance(centerx,centery,d_HF->d_xsize,0)) > md)
		md=d;
	if ((d=MathTrig::distance(centerx,centery,0,d_HF->d_ysize)) > md)
		md=d;
	if ((d=MathTrig::distance(centerx,centery,d_HF->d_xsize,d_HF->d_ysize)) > md)
		md=d;

	if (!scaleFullDist)
		md = (maxdist>md ? md-mindist : maxdist-mindist);

	for (int x=0; x<d_HF->d_xsize; x++)
		for (int y=0; y<d_HF->d_ysize; y++)
			{
			d=MathTrig::distance (centerx, centery, x, y);

			if (d >= mindist && d <= maxdist)
				{
				if (scale_factor >= 1)
					p=(d/md)*scale_factor;	// 1 .. scale_factor
				else
					p=(d/md)*(1-scale_factor);  // 0 .. 1
				d_HF->d_hf[d_HF->Pos2Off(x,y)]*=p;
				}
			else
			if (smoothFactor > 0 && d<mindist && d>smoothFloor)
				{
				PTYPE	pt=0;

				p = (mindist-d)/(mindist-smoothFloor);
				pt += d_HF->El(x,y)*(p*100);
				pt += d_HF->El(x,y)*((1-p)*100);
				d_HF->d_hf[d_HF->Pos2Off(x,y)] = pt/100;
				}
			else
			if (smoothFactor > 0 && d>maxdist && d<smoothCeil)
				{
				PTYPE	pt=0;

				p = (d-maxdist)/(maxdist-smoothCeil);
				pt += d_HF->El(x,y)*(p*100);
				pt += d_HF->El(x,y)*((1-p)*100);
				d_HF->d_hf[d_HF->Pos2Off(x,y)] = pt/100;
				}
			//printf ("%f, %f\n", p, d_HF->d_hf[d_HF->Pos2Off(x,y)]);
			//fflush (stdout);
			}

	d_HF->gatherStatistics ();
	d_HF->setSaved (FALSE);
	return (0);
}


/*
 *  mirror: mirror the height field on one of it's axis.
 *  Type can be 1, 2, 3, 4 for x-axis, y-axis, top (left
 *  to bottom right) or bottom (left to top right)
 */
int HeightFieldOps::mirror (int type)
{
	int	mid, x, y;
	PTYPE	t;
	

	// Mirror around X axis
	if (type == 1)
		{
		GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, 
			"Mirroring around X axis ...\n");

		mid = d_HF->d_ysize/2;
		for (y=0; y<mid; y++)
		   for (x=0; x<d_HF->d_xsize; x++)
			{
			t = d_HF->d_hf[d_HF->Pos2Off(x,y)];
			d_HF->d_hf[d_HF->Pos2Off(x,y)] = d_HF->d_hf[d_HF->Pos2Off(x,d_HF->d_ysize-y-1)];
			d_HF->d_hf[d_HF->Pos2Off(x,d_HF->d_ysize-y-1)] = t;
			}
		}
	else
	// Mirror around Y axis
	if (type == 2)
		{
		GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, 
			"Mirroring around Y axis ...\n");

		mid = d_HF->d_xsize/2;
		for (x=0; x<mid; x++)
		    for (y=0; y<d_HF->d_ysize; y++)
			{
			t = d_HF->d_hf[d_HF->Pos2Off(x,y)];
			d_HF->d_hf[d_HF->Pos2Off(x,y)] = d_HF->d_hf[d_HF->Pos2Off(d_HF->d_xsize-x-1,y)];
			d_HF->d_hf[d_HF->Pos2Off(d_HF->d_xsize-x-1,y)] = t;
			}
		}
	else
	// Mirror diagonally: this code will only work on square height fields
	if (type == 3 || type == 4)
		{
		if (d_HF->d_xsize != d_HF->d_ysize)
			{
			printf ("HeightFieldOps::mirror(): HF is not square ... cancelling\n");
			return (-1);
			}

		int 	i, lim, side = d_HF->d_xsize;

		if (type == 3)
		    {
		    GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, 
			"Mirroring around TL-BR axis ...\n");
		    for (i=0; i<side; i++)
			{
			for (x=i; x<side; x++)
				{
				t = d_HF->d_hf[d_HF->Pos2Off(x,i)];
				d_HF->d_hf[d_HF->Pos2Off(x,i)] = d_HF->d_hf[d_HF->Pos2Off(i,x)];
				d_HF->d_hf[d_HF->Pos2Off(i,x)] = t;
				}
			}
		    }
		else
		    {
		    GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, 
			"Mirroring around BL-TR axis ...\n");
		    for (i=side-1; i>=0; i--)
			{
			lim = side-i;
		        for (x=0; x<lim; x++)
				{
				t = d_HF->d_hf[d_HF->Pos2Off(x,i)];
				d_HF->d_hf[d_HF->Pos2Off(x,i)] = d_HF->d_hf[d_HF->Pos2Off(side-i,side-x)];
				d_HF->d_hf[d_HF->Pos2Off(side-i,side-x)] = t;
				}
			}
		    }
		}
	else
		SanityCheck::bailout ((1), "Uncaught mirror type operation requested", "HeightFieldOps::mirror");

	return (0);
}


/*
 *  fill: fill the Height Field up to elv
 */
int HeightFieldOps::fill (PTYPE	elv, float factor, bool keepMin)
{
	int	n=0;
	float	p;
	char	buf[80];

	if (!d_HF->d_hf)
		{
		printf ("Warning: hf==NULL in HeightFieldOps::Fill ... operation cancelled\n");
		return (-1);
		}
	if (elv < 0 || elv > 1)
		{
		cerr << "Error: elevation out of range (" << 
			elv << ") in HeightFieldOps::Fill() ...\n";
		return (-1);
		}

	sprintf (buf, "Filling to elevation %f, delta %f%% ... ", elv, (1-factor)*100);
	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, buf);

	if (keepMin)
		{
		for (int i=0; i<d_HF->d_size; i++)
			if (d_HF->d_hf[i]<elv && d_HF->d_hf[i]>d_HF->d_min)
				{
				p = elv-d_HF->d_hf[i];
				p -= (p*factor);
				d_HF->d_hf[i] = elv - p;
				n++;
				}
	    	}
	else
		{
		for (int i=0; i<d_HF->d_size; i++)
			if (d_HF->d_hf[i]<elv)
				{
				p = elv-d_HF->d_hf[i];
				p -= (p*factor);
				d_HF->d_hf[i] = elv - p;
				n++;
				}
	   	}

	sprintf (buf, "filled %d points\n", n);
	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, buf);

	d_HF->gatherStatistics ();
	d_HF->setSaved (FALSE);
	return (0);
}


