/* Terraform - (C) 1997-1998 Robert Gasch (Robert_Gasch@peoplesoft.com)
 *  - www.peoplesoft.com/peoplepages/g/robert_gasch/terraform/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#include "HeightFieldWriter.h"
#include "HeightFieldIO.h"
#include "GlobalTrace.h"
#include "GlobalSanityCheck.h" 



/*
 *  constructor: initialize
 */
HeightFieldWriter::HeightFieldWriter (HeightField *HF, FileIO *Fio) 
{
	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, "+++ HeightFieldWriter\n");

	SanityCheck::bailout ((!HF), "HF==NULL", "HeightFieldWriter::HeightFieldWriter");
	SanityCheck::bailout ((!HF->d_hf), "HF->d_hf==NULL", "HeightFieldWriter::HeightFieldWriter");
	SanityCheck::bailout ((!Fio), "Fio==NULL", "HeightFieldWriter::HeightFieldWriter");

	d_HF = HF;
	d_Fio = Fio;
}


/*
 *  destructor: nothing to clean up 
 */
HeightFieldWriter::~HeightFieldWriter ()
{
	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, "--- HeightFieldWriter\n");
}


/*
 *  writePGM: write a height field to a PGM ASCII file
 */
int HeightFieldWriter::writePGM ()
{
	int 	cnt; 			// count of #s printed
	int 	ix,iy;
	int	dotinterval=1;
	long 	idx;
	FILE 	*fp;
	int 	maxheight = 65535; 	// maximum value of array 
	char	*fname;
	char	buf[512];

	/* *** open PGM file *** */
	fp = d_Fio->open ("wb");
	fname = d_Fio->getName ();
	if (fp == NULL) 
		{
		cerr << "Can't open PGM output file [" << fname << "] ...\n";
		return(-1);
		}

	sprintf (buf, "Writing file %s (%d, %d) type PGM\nwriting: ", 
		fname, d_HF->d_xsize, d_HF->d_ysize);
	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, buf);
	dotinterval=(int)(d_HF->d_size/50);

	fprintf (fp, "P2\n");
	fprintf (fp, "%d %d\n", d_HF->d_xsize, d_HF->d_ysize);
	fprintf (fp, "%d\n", maxheight);

	idx = 0;
	cnt = 0;
	for (iy = 0; iy<d_HF->d_ysize; iy++) 
		{
		for (ix = 0; ix<d_HF->d_xsize; ix++) 
			{
			fprintf(fp,"%d ",(int) (maxheight*d_HF->d_hf[idx++]));
			if (++cnt > 10) 
				{ 
				fprintf (fp,"\n"); 
				cnt=0; 
				}
			if (!(idx%dotinterval) && 
			    GlobalTrace::trace(GlobalTrace::TRACE_VERBOSE, "."))
				;
    			}
    		if (cnt != 0) 
			{ 
			fprintf(fp,"\n"); 
			cnt = 0;
			}
		}

	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, "\n");

	d_Fio->close ();
	d_HF->setSaved (TRUE);
	return(0);
}


/*
 *  writePG8: write a height field to a PGM binary file
 */
int HeightFieldWriter::writePG8 ()
{
	int 	ix,iy;
	int	dotinterval=1;
	long 	idx;
	FILE 	*fp;
	int 	maxheight = 255; 
	char 	*fname;
	char	buf[512];
						/* open input file */
	/* *** open PG8 file *** */
	fp = d_Fio->open ("wb");
	fname = d_Fio->getName ();
	if (fp == NULL) 
		{
		cerr << "Can't open PG8 output file [" << fname << "] ...\n";
		return(-1);
		}

	sprintf (buf, "Writing file %s (%d, %d) type PG8\nwriting: ", 
		fname, d_HF->d_xsize, d_HF->d_ysize);
	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, buf);
	dotinterval=(int)(d_HF->d_size/50);

	fprintf(fp, "P5\n");
	fprintf(fp, "%d %d\n", d_HF->d_xsize, d_HF->d_ysize);
	fprintf(fp, "%d\n", maxheight);

	idx = 0;
	for (iy = 0; iy<d_HF->d_ysize; iy++) 
		for (ix = 0; ix<d_HF->d_xsize; ix++) 
			{
			fprintf(fp,"%c",(BYTE)(255*d_HF->d_hf[idx++]));
			if (!(idx%dotinterval) && 
			    GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, "."))
				;
			}

	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, "\n");

	d_Fio->close ();
	d_HF->setSaved (TRUE);
	return(0);
} 


/*
 *  writeTGA: write the Height Field to a TGA file 
 */
int HeightFieldWriter::writeTGA ()
{
	int 		ix, iy, dotinterval=1;
	FILE 		*fp;
	char		*fname;
	double 		temp;

	unsigned char 	red, green, blue;
	unsigned char 	IDLength;             /* length of Identifier String */
	unsigned char 	CoMapType;            /* 0 = no map */
	unsigned char 	ImgType;              /* image type (see below for values) */
	unsigned char 	Index_lo, Index_hi;   /* index of first color map entry */
	unsigned char 	Length_lo, Length_hi; /* number of entries in color map */
	unsigned char 	CoSize;               /* size colormap entry (15,16,24,32) */
	unsigned char 	X_org_lo, X_org_hi;   /* x origin of image */
	unsigned char 	Y_org_lo, Y_org_hi;   /* y origin of image */
	unsigned char 	Width_lo, Width_hi;   /* width of image */
	unsigned char 	Height_lo, Height_hi; /* height of image */
	unsigned char 	PixelSize;            /* pixel size (8,16,24,32) */
	long 		idx=0;

	/* *** open TGA file *** */
	fp = d_Fio->open ("wb");
	fname = d_Fio->getName ();
	if (fp == NULL) 
		{
		cerr << "Can't open TGA output file [" << fname << "] ...\n";
		return(-1);
		}

	if (GlobalTrace::isSet (GlobalTrace::TRACE_VERBOSE))
		{
		printf ("Writing file %s (%d, %d) type TGA\nwriting: ", 
			fname, d_HF->d_xsize, d_HF->d_ysize);
		dotinterval=(int)(d_HF->d_size/50);
		}

	/* *** Write out data to TGA format file *** */
	/* *** eighteen-byte header *** */
	/* *** b0-b7 *** */
	IDLength = 0; CoMapType = 0; ImgType = 2; Index_lo = 0; Index_hi=0;
	Length_lo = 0; Length_hi = 0; CoSize = 0; X_org_lo = 0; X_org_hi = 0;
	temp = d_HF->d_xsize / 256;
	Width_hi = (unsigned char) temp;     /* low and high bytes of x width */
	Width_lo = d_HF->d_xsize % 256;
	temp = d_HF->d_ysize / 256;
	Height_hi = (unsigned char) temp;
	Height_lo = d_HF->d_ysize % 256;
	PixelSize = 24;
	Y_org_lo = 0;
	Y_org_hi = 0;

	fprintf(fp ,"%c",IDLength);
	fprintf(fp ,"%c",CoMapType);
	fprintf(fp ,"%c",ImgType);
	fprintf(fp ,"%c",Index_lo);
	fprintf(fp ,"%c", Index_hi);
	fprintf(fp ,"%c",Length_lo);
	fprintf(fp ,"%c",Length_hi);
	fprintf(fp ,"%c",CoSize);
	fprintf(fp ,"%c",X_org_lo); 
	fprintf(fp ,"%c",X_org_hi);
	fprintf(fp ,"%c",Y_org_lo); 
	fprintf(fp ,"%c",Y_org_hi);
	fprintf(fp ,"%c",Width_lo);
	fprintf(fp ,"%c",Width_hi);              /* file in lo-byte, hi-byte order b12,b13*/
	fprintf(fp ,"%c",Height_lo);
	fprintf(fp ,"%c",Height_hi);    /* ysize b14, b15 */
	fprintf(fp ,"%c",PixelSize);
	fprintf(fp ,"%c",(unsigned char) 0x20);        /* descriptor byte b17 */

	blue = 0;
	for (iy = 0; iy < d_HF->d_ysize; iy++) 
		for (ix = 0; ix < d_HF->d_xsize; ix++) 
			{
			red = (UI)(65535.0*d_HF->El(ix,iy))/256; /* high byte */
			green = (UI)(65535.0*d_HF->El(ix,iy)) % 256; /* low byte */
			fprintf(fp ,"%c%c%c",blue,green,red);

			if (!(++idx%dotinterval) &&
			    GlobalTrace::trace(GlobalTrace::TRACE_VERBOSE, "."))
				;
			}

	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, "\n");

	d_Fio->close ();
	d_HF->setSaved (TRUE);
	return(0);
} 



/*
 * writeMAT: write out data matrix in Matlab Level 1.0 format
 * 	     creates data header fhdr from xsize, ysize and
 *	     writes data array 'hf'
 */

int HeightFieldWriter::writeMAT ()
{
	FILE 	*fp;
	Fmatrix fhdr;
	int 	rv,mn,i,j,c;
	char	*fname;
	char	*pname = "topo";
	int 	dotinterval=1;
	long	idx=0;

	/* *** open MAT file *** */
	fp = d_Fio->open ("wb");
	fname = d_Fio->getName ();
	if (fp == NULL) 
		{
		cerr << "Can't open MAT output file [" << fname << "] ...\n";
		return(-1);
		}

	if (GlobalTrace::isSet (GlobalTrace::TRACE_VERBOSE))
		{
		printf ("Writing file %s (%d, %d) type MAT\nwriting: ", 
			fname, d_HF->d_xsize, d_HF->d_ysize);
		dotinterval=(int)(d_HF->d_size/50);
		}

	fhdr.type = NUM_FMT*1000 + PREC*10;
	fhdr.ncols = d_HF->d_xsize;
	fhdr.mrows = d_HF->d_ysize;
	fhdr.imagf = FALSE;
	fhdr.namelen = strlen(pname) + 1;

	// write header 
	rv = fwrite (&fhdr, sizeof(Fmatrix),(size_t)1, fp); 
	if (rv != 1) 
		{
		perror("Error writing file.\n");
 		return(1);
		}

	// write mx name 
	rv = fwrite(pname, sizeof(char), (size_t)fhdr.namelen, fp);   
	if (rv != fhdr.namelen) 
		{
		perror("Error writing file.\n");
		return(1);
		}
	mn = fhdr.mrows * fhdr.ncols;

	for (i=0; i<d_HF->d_xsize; i++) 
		for (j=0; j<d_HF->d_ysize; j++) 
			{   
       			c = fwrite (&(d_HF->d_hf[d_HF->Pos2Off(i,j)]), sizeof(PTYPE), 1, fp);
       			if (c != 1) 
				{
				perror ("Trouble writing data to file.\n");
				return(1);
				}
			if (!(++idx%dotinterval) &&
			    GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, "."))
				;
			} 

	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, "\n");

	if (fclose( fp ) == EOF) 
		{
		perror("Error closing file.\n");
		return(1);
		}

	d_Fio->close ();
	d_HF->setSaved (TRUE);
	return(0);
} 


int HeightFieldWriter::writeOCT ()
{
	int 	ix,iy;
	long 	idx;
	FILE 	*fp;
	char 	*fname,
		*mname = "topo";
	int	dotinterval=1;

	/* *** open OCT file *** */
	fp = d_Fio->open ("wb");
	fname = d_Fio->getName ();
	if (fp == NULL) 
		{
		cerr << "Can't open OCT output file [" << fname << "] ...\n";
		return(-1);
		}

	if (GlobalTrace::isSet (GlobalTrace::TRACE_VERBOSE))
		{
		printf ("Writing file %s (%d, %d) type OCT\nwriting: ", 
			fname, d_HF->d_xsize, d_HF->d_ysize);
		dotinterval=(int)(d_HF->d_size/50);
		}

	fprintf (fp,"# name: %s\n", mname);
	fprintf (fp,"# type: matrix\n");
	fprintf (fp,"# rows: %d\n", d_HF->d_ysize);
	fprintf (fp,"# columns: %d\n", d_HF->d_xsize);
  
	idx = 0;
	for (iy = 0; iy<d_HF->d_ysize; iy++) 
		{
		for (ix = 0; ix<d_HF->d_xsize; ix++) 
			{
			fprintf (fp,"%e ", d_HF->d_hf[idx++]);
			if (!(idx%dotinterval) &&
			    GlobalTrace::trace(GlobalTrace::TRACE_VERBOSE, "."))
				;
			}
		fprintf(fp,"\n");
		}

	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, "\n");

	d_Fio->close ();
	d_HF->setSaved (TRUE);
	return 0;
}

