/* Terraform - (C) 1997-1998 Robert Gasch (Robert_Gasch@peoplesoft.com)
 *  - www.peoplesoft.com/peoplepages/g/robert_gasch/terraform/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#include <stdlib.h>
#include "TFDialogRoughSmooth.h"
#include "GlobalSanityCheck.h"
#include "GlobalTrace.h"


/*
 *  constructor: initialize all data members 
 */
TFDialogRoughSmooth::TFDialogRoughSmooth (bool roughen, HeightField *HF, HeightFieldDraw *HFD, char *title)
	        : GuiDialogOAC (title, title), 
		  d_vBoxList (TRUE, 5),
		  d_hBoxRSFactor (TRUE, 5),
		  d_hBoxBigGrid (TRUE, 5),
		  d_adjRSFactor (1, 0, 1, 0.01),
		  d_hsRSFactor (d_adjRSFactor),
		  d_lblRSFactor ("Adjustment Factor"),
		  d_lblBigGrid ("Big Sampling Grid")
{
	GlobalTrace::trace (GlobalTrace::TRACE_FLOW, "+++ TFDialogRoughSmooth\n");

	d_HFO = NULL;
	buildDialogWindow ();
	setHFobjs (roughen, HF, HFD, title);
}


/*
 *  destructor: clean up 
 */
TFDialogRoughSmooth::~TFDialogRoughSmooth ()
{
	if (d_HFO)
		delete d_HFO;

	GlobalTrace::trace (GlobalTrace::TRACE_FLOW, "--- TFDialogRoughSmooth\n");
}


/*
 *  setHFobjs: set the HF objects we operate on
 */
void TFDialogRoughSmooth::setHFobjs (bool roughen, HeightField *HF, HeightFieldDraw *HFD, char *title)
{
	char 	*oldTitle;

	SanityCheck::bailout ((!HF), "HF==NULL", "TFDialogRoughSmooth::setHFobjs");
	SanityCheck::bailout ((!HFD), "HFD==NULL", "TFDialogRoughSmooth::setHFobjs");
	SanityCheck::bailout ((!title), "title==NULL", "TFDialogRoughSmooth::setHFobjs");
	SanityCheck::bailout ((!this->d_baseFrame), "TFDialogRoughSmooth called without Frame label",
		"TFDialogRoughSmooth::setHFobjs");

	if (d_HFO)
		delete d_HFO;

	d_HF = HF;
	d_HFD = HFD;
	d_HFO = new HeightFieldOps (d_HF);
	d_roughen = roughen;
	this->set_title (title);

	oldTitle = this->d_frameTitle;
	if (d_roughen)
		this->d_frameTitle = strdup ("Roughen height field"); 
	else
		this->d_frameTitle = strdup ("Smooth height field"); 
	this->d_baseFrame->set_label ();
	this->d_baseFrame->set_label (this->d_frameTitle);
	free (oldTitle);
}


/*
 *  fillVBox: fill the dialog's VBox
 */
void TFDialogRoughSmooth::fillVBox ()
{
	SanityCheck::bailout ((!this->d_baseFrame), "TFDialogRoughSmooth called without Frame label",
		"TFDialogRoughSmooth::fillVBox");

	d_hBoxRSFactor.pack_start (d_lblRSFactor, TRUE, TRUE, s_HBOff);
	d_lblRSFactor.show ();
	d_hsRSFactor.set_update_policy (GTK_UPDATE_DELAYED);
	d_hsRSFactor.set_digits (2);
	d_hsRSFactor.set_draw_value (TRUE);
	d_hBoxRSFactor.pack_end (d_hsRSFactor, TRUE, TRUE, s_HBOff);
	d_hsRSFactor.show ();
	d_vBoxList.pack_start (d_hBoxRSFactor, TRUE, TRUE, s_VBOff);
	d_hBoxRSFactor.show ();

	d_hBoxBigGrid.pack_start (d_lblBigGrid, TRUE, TRUE, s_HBOff);
	d_lblBigGrid.show ();
	d_btnBigGrid.set_active (FALSE);
	d_hBoxBigGrid.pack_end (d_btnBigGrid, TRUE, TRUE, s_HBOff);
	d_btnBigGrid.show ();
	d_vBoxList.pack_start (d_hBoxBigGrid, TRUE, TRUE, s_VBOff);
	d_hBoxBigGrid.show ();

	this->d_baseFrame->add (d_vBoxList);	
	d_vBoxList.show ();
	this->d_baseFrame->show ();
	this->get_vbox()->show ();
}



void TFDialogRoughSmooth::buildDialogWindow ()
{
	this->set_usize (250, 150);
	this->set_title (this->d_winTitle);

	this->get_vbox()->set_border_width (2);

	this->setupVBox ();
	fillVBox ();
	this->fillActionArea ();
	
	this->show ();
}


void TFDialogRoughSmooth::buttonCallbackOK ()
{	
	this->hide ();
	gdk_flush ();
	buttonCallbackApply ();
}


/*
 *  buttonCallbackApply: transform the height field
 */
void TFDialogRoughSmooth::buttonCallbackApply ()
{
	bool			bigGrid=FALSE;
	float			value; 

	value = d_adjRSFactor.get_value (); 
	bigGrid = d_btnBigGrid.get_active ();

	SanityCheck::bailout ((!d_HFO), "d_HFO==NULL", "TFDialogRoughSmooth::buttonCallbackApply");

	d_HFO->roughSmooth (d_roughen, bigGrid, value);
	d_HFD->draw ();
}


gint TFDialogRoughSmooth::delete_event_impl (GdkEventAny *)
{
	this->hide ();
	return (0);
}


/*
int GlobalTrace::d_traceLevel=255;    // instantiate static data member
int main (int argc, char *argv[])
{
	Gtk_Main        gtkMain (&argc, &argv);
	TFDialogRoughSmooth *testDialog;

	testDialog = new TFDialogRoughSmooth (NULL, NULL, "TestTitle", "TestMessage");
	testDialog->show ();
	gtkMain.run ();
}
*/
