/* Terraform - (C) 1997-1998 Robert Gasch (Robert_Gasch@peoplesoft.com)
 *  - www.peoplesoft.com/peoplepages/g/robert_gasch/terraform/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#include <math.h>
#include "terraform.h"
#include "TFGui.h"
#include "FileIO.h"
#include "GlobalSanityCheck.h"
#include "GlobalTrace.h"
#include "GuiColormap.h"
#include "GuiDialogPreview.h"
#include "GuiDialogReallyQuit.h"
#include "HeightFieldGenRandom.h"
#include "HeightFieldIO.h"
#include "TFDialogPreview.h"
#include "TFDialogErode.h"
#include "TFDialogFill.h"
#include "TFDialogFlowmap.h"
#include "TFDialogFold.h"
#include "TFDialogGenSSynth.h"
#include "TFDialogGenSubdiv.h"
#include "TFDialogLinearScale.h"
#include "TFDialogMirror.h"
#include "TFDialogOptions.h"
#include "TFDialogPPov.h"
#include "TFDialogRescale.h"
#include "TFDialogRotate.h"
#include "TFDialogRoughSmooth.h"
#include "TFDialogTransform.h"
#include "Timer.h"
#include "MathRandom.h"

#ifdef HAVE_IMLIB
#include "GuiDialogAboutDissolve.h"
#endif


extern int readhf (HeightField **HF, char *fname);

static int		cs_fileselMode;	// flag for file selection pointer


#define LOAD		0
#define SAVE		1
#define BANDSIZE	15


/*
 *  constructor: initialize everything 
 */
TFGui::TFGui (Gtk_Main *gtkMain, gint argc, char *argv[]) 
{
	d_HF = NULL;
	d_HFD = NULL;
	d_fileSelection = NULL;
	d_gtkMain = gtkMain;
	d_optionsDialog = NULL;
	d_erodeDialog=d_genSSynthDialog=d_genSubdivDialog=d_fillDialog=
		d_placePovDialog= d_mirrorDialog=d_flowmapDialog=d_rescaleDialog=
		d_rotateDialog=d_roughSmoothDialog=d_transformDialog = NULL;
	d_aboutDialog = NULL;
	d_quitDialog = NULL;
	d_fastWire = FALSE;
	d_rotate = FALSE;
	d_autoRotate = FALSE;
	
	//connect_to_method (this->delete_event, this, &deleteCallback);

	activateMenus (FALSE);
	gdk_flush ();
}


gint TFGui::idleRedrawCallback ()
{
	GlobalTrace::trace (GlobalTrace::TRACE_FLOW, "Idle Redraw ...");
	drawHeightField ();
	return 0;
}


gint TFGui::autoRotateCallback ()
{
	GlobalTrace::trace (GlobalTrace::TRACE_FLOW, "Auto Rotate ...");

	if (d_HFD->rotateFastView (0.0, (M_PI/180)*1, 0.0))
		{
		// !!!!!!!!! set incorrectly selected check box back to FALSE
		d_autoRotate = FALSE;
		}
	else
		drawHeightField ();

	return d_autoRotate;
}


gint TFGui::calibrateAutoRotate (int nSamples)
{
	Timer		timer;
	long		eusec;		// elapsed microseconds
	struct timeval  *etime;		// elapsed time

	GlobalTrace::trace (GlobalTrace::TRACE_FLOW, "Calibrating AutoRotate ...");

	timer.start ();
	for (int i=0; i<nSamples; i++)
		{
		autoRotateCallback ();
		gdk_flush ();
		}
	timer.stop ();
	etime = timer.getElapsedTime ();
	eusec = etime->tv_usec + etime->tv_sec*1000000;
	eusec /= nSamples;

	GlobalTrace::trace (GlobalTrace::TRACE_FLOW,  " done\n");

	return eusec;
}


/*
 *  set the Height Field we draw 
 */
void TFGui::setHeightField (HeightField *HF)
{
	int	x, y;
	SanityCheck::bailout ((!HF), "HF==NULL", "TFGui::setHeightField");

	if (d_HF && d_HF!=HF)
		delete d_HF;
	d_HF=HF;

	if (d_HFD)
		delete d_HFD;
	d_HFD=new HeightFieldDraw (d_HF, this);
	d_HFD->setScale (1.0, &x, &y);

	if (this->d_useLinearColormap)
		d_HFD->setColormap (this->d_linearColorMap);
	else
		d_HFD->setColormap (this->d_bandColorMap, 
				    this->d_bandColorMap->getNumBands(),
				    this->d_bandColorMap->getBandSize(), 4);

	if (GlobalTrace::isSet(GlobalTrace::TRACE_DEBUG))
		d_HFD->printSettings();

	//d_HFD->draw ();
	connect_to_method (Gtk_Main::timeout(50), this, &idleRedrawCallback);
	activateMenus (TRUE);
}


/*
 *  drawHeightField: draw and then reset fore- and background colors
 */
void TFGui::drawHeightField ()
{
	SanityCheck::bailout ((!d_HF), "d_HF==NULL", "TFGui::drawHeightField");
	SanityCheck::bailout ((!d_HFD), "d_HFD==NULL", "TFGui::drawHeightField");

	d_HFD->draw ();
	this->setColors (&(this->d_linearColorMap->d_black));
}



/*
 *  fileMenuCallback: process a menu choice in the file menu
 */
void TFGui::fileMenuCallback (string m)
{
	int		rc=0;
	char		*menuitem = const_cast<char*>(m.c_str()),
			*aboutText[] = { "brought to you by", 
					"Robert_Gasch@peoplesoft.com", };

	if (!strcmp (menuitem, MENU_FILE_NEW_SYNTHESIS))
		{
		if (!d_genSSynthDialog)
			d_genSSynthDialog = new TFDialogGenSSynth (new HeightField(), this);
		else
			{
			if (d_HF)
				(dynamic_cast<TFDialogGenSSynth*>(d_genSSynthDialog))->setHFobjs (d_HF, this);
			else
				(dynamic_cast<TFDialogGenSSynth*>(d_genSSynthDialog))->setHFobjs (new HeightField(), this);
			}
		
		d_genSSynthDialog->show ();
		}
	else
	if (!strcmp (menuitem, MENU_FILE_NEW_SUBDIVIDE))
		{
		if (!d_genSubdivDialog)
			d_genSubdivDialog = new TFDialogGenSubdiv (new HeightField(), this); 
		else
			{
			if (d_HF)
				(dynamic_cast<TFDialogGenSubdiv*>(d_genSubdivDialog))->setHFobjs (d_HF, this);
			else
				(dynamic_cast<TFDialogGenSubdiv*>(d_genSubdivDialog))->setHFobjs (new HeightField(), this);
			}
		
		d_genSubdivDialog->show ();
		}
	else
	if (!strcmp (menuitem, MENU_FILE_NEW_RANDOM))
		{
		if (!d_HF)
			d_HF = new HeightField ();

		HeightFieldGenRandom::generate (d_HF, this->width());
		d_HFD->draw ();
		}
	else
	if (!strcmp (menuitem, MENU_FILE_OPEN))
		{
		cs_fileselMode=LOAD;			// 0 = load, 1 = save
		fileSelection ();
		}
	else
	if (!strcmp (menuitem, MENU_FILE_RELOAD))
		{
		if (d_HF && !d_HF->isSaved())
			{
			if (!d_quitDialog)
				d_quitDialog = new GuiDialogReallyQuit ("", "");
			else
				d_quitDialogConnection.disconnect ();

			d_quitDialog->set_title (d_HF->getName());
			d_quitDialog->setLabel ("Unsaved Changes! Reload Anyways?");
			d_quitDialogConnection = connect_to_method (
				d_quitDialog->getYesButton()->clicked, 
				this, &reloadCallback);
			d_quitDialog->show ();
			}
		else
		if (d_HF)
			reloadCallback ();
		}
	else
	if (!strcmp (menuitem, MENU_FILE_SAVE))
		{
		if (d_HF)
			{
			HeightFieldIO		*HFIO;

			HFIO = new HeightFieldIO (d_HF->getName());
			rc = HFIO->write (d_HF);
			}
		}
	else
	if (!strcmp (menuitem, MENU_FILE_SAVEAS))
		{
		if (d_HF)
			{
			cs_fileselMode=SAVE;			// 0 = load, 1 = save
			fileSelection ();
			}
		}
	else
	if (!strcmp (menuitem, MENU_FILE_CLOSE))
		{
		if (d_HF && !d_HF->isSaved())
			{
			if (!d_quitDialog)
				d_quitDialog = new GuiDialogReallyQuit ("", "");
			else
				d_quitDialogConnection.disconnect ();

			d_quitDialog->set_title (d_HF->getName());
			d_quitDialog->setLabel ("Unsaved Changes! Close Anyways?");
			d_quitDialogConnection = connect_to_method 
				(d_quitDialog->getYesButton()->clicked, 
				 this, &closeCallback);
			d_quitDialog->show ();
			}
		else
		if (d_HF)
			closeCallback ();
		}
	else
	if (!strcmp (menuitem, MENU_FILE_ABOUT))
		{
		if (!d_aboutDialog)
			{
#ifdef HAVE_IMLIB
			char	buf[256];
			sprintf (buf, "%s/terraform-logo.jpg", TF_DATADIR);
			FileIO	*fio = new FileIO (buf);
			if (fio->exists())
				d_aboutDialog = new GuiDialogAboutDissolve ("About Terraform", buf, aboutText, 2);
			else
				fprintf (stderr, "Error opening %s\n", buf);
			delete fio;
#else
			d_aboutDialog = new GuiDialogAbout ("About Terraform", NULL, aboutText, 2);
#endif
			if (d_aboutDialog)
				d_aboutDialog->buildDialogWindow ();
			}

		if (d_aboutDialog)
			d_aboutDialog->show ();
		}
	else
	if (!strcmp (menuitem, MENU_FILE_OPTIONS))
		{
		if (d_HF)
			{
			if (!d_optionsDialog)
				d_optionsDialog = new TFDialogOptions (d_HF, d_HFD);
			else
				d_optionsDialog->show ();
			}
		}
	else
	if (!strcmp (menuitem, MENU_FILE_EXIT))
		{
		if (d_HF && !d_HF->isSaved())
			{
			if (!d_quitDialog)
				d_quitDialog = new GuiDialogReallyQuit ("", "");
			else
				d_quitDialogConnection.disconnect ();

			d_quitDialog->set_title (d_HF->getName());
			d_quitDialog->setLabel ("Unsaved Changes! Exit anyway?");
			d_quitDialogConnection = connect_to_method (
				d_quitDialog->getYesButton()->clicked, 
				this, &destroyCallback);
			d_quitDialog->show ();
			}
		else
			destroyCallback ();
		}
	else
		SanityCheck::bailout (TRUE, "uncaught <file> menuitem", "TFGui::fileMenuCallback");

	if (rc)
		g_print ("Warning: last command returned an Error (%d)\n", rc);
	fflush (stdout);
}


/*
 *  hfMenuCallback: process a menu choice in the height field menu
 */
void TFGui::hfMenuCallback (string m)
{
	char		*menuitem = const_cast<char*>(m.c_str());

	if (!d_HF)
		return;

	if (!strcmp (menuitem, MENU_HF_FILL))
		{
/*
		if (!d_fillDialog) 
			d_fillDialog = new TFDialogFill (d_HF, d_HFD);
		else
			(dynamic_cast<TFDialogFill*>(d_fillDialog))->setHFobjs (d_HF, d_HFD);
		d_fillDialog->show ();
*/
		if (!d_previewDialog) 
			d_previewDialog = new TFDialogPreview (d_HF, d_HFD);
		d_previewDialog->show ();
		}
	else
	if (!strcmp (menuitem, MENU_HF_FLOWMAP))
		{
		if (!d_flowmapDialog) 
			d_flowmapDialog = new TFDialogFlowmap (d_HF, d_HFD);
		else
			(dynamic_cast<TFDialogFlowmap*>(d_flowmapDialog))->setHFobjs (d_HF, d_HFD);

		d_flowmapDialog->show ();
		}
	else
	if (!strcmp (menuitem, MENU_HF_FOLD))
		{
		if (!d_foldDialog)
			d_foldDialog = new TFDialogFold (d_HF, d_HFD);
		else
			(dynamic_cast<TFDialogFold*>(d_foldDialog))->setHFobjs (d_HF, d_HFD);

		d_foldDialog->show ();
		}
	else
	if (!strcmp (menuitem, MENU_HF_INVERT))
		{
		if (d_HF)
			{
			HeightFieldOps	*HFO;

			HFO = new HeightFieldOps (d_HF);
			HFO->invert ();
			d_HFD->draw ();
			delete HFO;
			}
		}
	else
	if (!strcmp (menuitem, MENU_HF_LINSCALE))
		{
		if (!d_linearScaleDialog) 
			d_linearScaleDialog = new TFDialogLinearScale (d_HF, d_HFD);
		else
			(dynamic_cast<TFDialogLinearScale*>(d_linearScaleDialog))->setHFobjs (d_HF, d_HFD);

		d_linearScaleDialog->show ();
		}
	else
	if (!strcmp (menuitem, MENU_HF_MIRROR))
		{
		if (!d_mirrorDialog)
			d_mirrorDialog = new TFDialogMirror (d_HF, d_HFD);
		else
			(dynamic_cast<TFDialogMirror*>(d_mirrorDialog))->setHFobjs (d_HF, d_HFD);

		d_mirrorDialog->show ();
		}
	else
	if (!strcmp (menuitem, MENU_HF_RESCALE))
		{
		if (!d_rescaleDialog) 
			d_rescaleDialog = new TFDialogRescale (d_HF, d_HFD);
		else
			(dynamic_cast<TFDialogRescale*>(d_rescaleDialog))->setHFobjs (d_HF, d_HFD);

		d_rescaleDialog->show ();
		}
	else
	if (!strcmp (menuitem, MENU_HF_ROTATE))
		{
		if (!d_rotateDialog) 
			d_rotateDialog = new TFDialogRotate (d_HF, d_HFD);
		else
			(dynamic_cast<TFDialogRotate*>(d_rotateDialog))->setHFobjs (d_HF, d_HFD);

		d_rotateDialog->show ();
		}
	else
	if (!strcmp (menuitem, MENU_HF_ROUGHEN))
		{
		if (!d_roughSmoothDialog) 
			d_roughSmoothDialog = new TFDialogRoughSmooth (TRUE, d_HF, d_HFD, "Terraform Roughen Dialog");
		else
			(dynamic_cast<TFDialogRoughSmooth*>(d_roughSmoothDialog))->setHFobjs (TRUE, d_HF, d_HFD, "Terraform Roughen Dialog");

		d_roughSmoothDialog->show ();
		}
	else
	if (!strcmp (menuitem, MENU_HF_SMOOTH))
		{
		if (!d_roughSmoothDialog) 
			d_roughSmoothDialog = new TFDialogRoughSmooth (FALSE, d_HF, d_HFD, "Terraform Smooth Dialog");
		else
			(dynamic_cast<TFDialogRoughSmooth*>(d_roughSmoothDialog))->setHFobjs (FALSE, d_HF, d_HFD, "Terraform Smooth Dialog");

		d_roughSmoothDialog->show ();
		}
	else
	if (!strcmp (menuitem, MENU_HF_TRANSFORM))
		{
		if (!d_transformDialog)
			d_transformDialog = new TFDialogTransform (d_HF, d_HFD);
		else
			(dynamic_cast<TFDialogTransform*>(d_transformDialog))->setHFobjs (d_HF, d_HFD);

		d_transformDialog->show ();
		}
	else
	if (!strcmp (menuitem, MENU_HF_ERODE))
		{
		//HFO->erode (1000, 200, 0.000043, TRUE, TRUE);
		if (!d_erodeDialog)
			d_erodeDialog = new TFDialogErode (d_HF, d_HFD);
		else
			(dynamic_cast<TFDialogErode*>(d_erodeDialog))->setHFobjs (d_HF, d_HFD);

		d_erodeDialog->show ();
		}
	else
	if (!strcmp (menuitem, MENU_HF_PLACEPOV))
		{
		if (!d_placePovDialog)
			d_placePovDialog = new TFDialogPPov (d_HF);
		else
			(dynamic_cast<TFDialogPPov*>(d_placePovDialog))->setHF (d_HF);

		d_placePovDialog->show ();
		}
	else
		SanityCheck::bailout (TRUE, "uncaught <hf> menuitem", "TFGui::fileMenuCallback");

	fflush (stdout);
}


/*
 *  viewMenuCallback: process a menu choice in the view menu
 */
void TFGui::viewMenuCallback (string m)
{
	char		*menuitem = const_cast<char*>(m.c_str());
	long		delay;

	if (!d_HFD)
		return ;

	if (!strcmp (menuitem, MENU_VIEW_FASTWIRE))
		{
		Gtk_Widget		*w;

		if (d_fastWire)
			d_fastWire = FALSE;
		else
			d_fastWire = TRUE;
		w = (Gtk_Widget *)this->d_menuHash->lookup(MENU_VIEW_AUTOROTATE);
		w->set_sensitive (d_fastWire);
		d_HFD->setFastWire (d_fastWire);
		if (d_HFD->getMode()==WIRE)
			drawHeightField ();
		}
	else
	if (!strcmp (menuitem, MENU_VIEW_AUTOROTATE))
		{
		if (d_autoRotate)
			d_autoRotate = FALSE;
		else
			{
			d_autoRotate = TRUE;
			delay = (long int)(calibrateAutoRotate(10)/1000*1.1);
			connect_to_method (Gtk_Main::timeout(delay), this, &autoRotateCallback);
			}
		}
	else
	if (!strcmp (menuitem, MENU_VIEW_2DPLANE) ||
	    !strcmp (menuitem, MENU_VIEW_3DWIRE)  ||
	    !strcmp (menuitem, MENU_VIEW_3DHEIGHT)||
	    !strcmp (menuitem, MENU_VIEW_3DLIGHT) )
		{
		if (!d_HFD->setMode (menuitem))
			drawHeightField ();
		}
	else
		SanityCheck::bailout (TRUE, "uncaught <view> menuitem", "TFGui::fileMenuCallback");

	fflush (stdout);
}

/*
 *  colormapMenuCallback: process a menu choice in the colormap menu
 */
void TFGui::colormapMenuCallback (string m)
{
	char		*menuitem = const_cast<char*>(m.c_str());

	if (!strcmp (menuitem, MENU_CMAP_GRAYSCALE))
		{
		this->d_useLinearColormap = TRUE;
		if (d_HFD)
			{
			if (!d_HFD->setColormap (this->d_linearColorMap))
				d_HFD->draw ();
			}
		}
	else
	if (!strcmp (menuitem, MENU_CMAP_COLORBANDS))
		{
		this->d_useLinearColormap = FALSE;
		if (d_HFD)
			{
			if (!d_HFD->setColormap (this->d_bandColorMap, 
			     this->d_bandColorMap->getNumBands(),
			     this->d_bandColorMap->getBandSize(), 4))
				d_HFD->draw ();
			}
		}
	else
	if (!strcmp (menuitem, MENU_CMAP_DESERT))
		{
		this->d_useLinearColormap = FALSE;
		if (d_HFD)
			{
			if (!d_HFD->setColormap (this->d_bandColorMap2, 
			     this->d_bandColorMap2->getNumBands(),
			     this->d_bandColorMap2->getBandSize(), 1))
				d_HFD->draw ();
			}
		}
	else
	if (!strcmp (menuitem, MENU_CMAP_REDHOT))
		{
		this->d_useLinearColormap = TRUE;
		if (d_HFD)
			{
			if (!d_HFD->setColormap (this->d_linearColorMap2))
				d_HFD->draw ();
			}
		}
	else
		SanityCheck::bailout (TRUE, "uncaught <colormap> menuitem", "TFGui::fileMenuCallback");

	fflush (stdout);
}


/*
 *  destroyCallback: handle a destroy event 
 */
void TFGui::destroyCallback ()
{
	GlobalTrace::trace (GlobalTrace::TRACE_FLOW, "Entering destroyCallback\n"); 

	closeCallback ();

	g_print ("Shutting Down ...\n");
	gtk_exit (0);
	exit (0);
}


/*
 *  closeCallback: delete the height field
 */
void TFGui::closeCallback ()
{
	GlobalTrace::trace (GlobalTrace::TRACE_FLOW, "Entering closeCallback\n"); fflush (stdout);

	if (d_quitDialog)
		d_quitDialog->hide ();

	if (d_HFD)
		{
		if (this->d_useLinearColormap)
			d_HFD->clear ();
		else
			d_HFD->clear ();
		delete d_HFD;
		d_HFD=NULL;
		}
	if (d_HF)
		{
		delete d_HF;
		d_HF=NULL;
		}

	activateMenus (FALSE);

	if (this->d_useLinearColormap)
		this->setColors (&(d_linearColorMap->d_black));
	else
		this->setColors (&(d_bandColorMap->d_black));
	this->drawRectangle (0, 0, this->width(), 
			this->height(), TRUE);
	this->sync ();
}


/*
 *  reloadCallback: process a reload request 
 */
void TFGui::reloadCallback ()
{
	if (d_quitDialog)
		d_quitDialog->hide ();

	if (d_HF)
		{
		HeightField		*HF = NULL;

		readhf (&HF, d_HF->getName());
		setHeightField (HF);
		}
}


/*
 *  fileSelectionOK: user pressed OK in file selection dialog
 */
void TFGui::fileSelectionOK ()
{
	gchar 	*fname;
	int	rc;

	d_fileSelection->hide ();

	fname = (char *) (d_fileSelection->get_filename().data());
	//g_print ("%s\n", fname);

	//create_progress_bar (fname);

	if (cs_fileselMode == LOAD)
		{
		HeightField	*HF=NULL;

		rc = readhf (&HF, fname);
		setHeightField (HF);
		}
	else
		{
		HeightFieldIO		*HFIO;

		HFIO = new HeightFieldIO (fname);
                rc = HFIO->write (d_HF);
		delete HFIO;
		}
	if (rc)
		cout << "Warning: last command returned Error " << rc << "\n";

	activateMenus (TRUE);
}


/*
 *  fileSelectionCancel: user pressed Cancel in file selection dialog
 */
void TFGui::fileSelectionCancel ()
{
	d_fileSelection->set_filename ("");
	d_fileSelection->hide ();
}


/*
 *  fileSelection: open a file selection dialog
 */
void TFGui::fileSelection ()
{
	if (!d_fileSelection)
		{
		d_fileSelection = new Gtk_FileSelection ("Load Height Field");
		d_fileSelection->set_position (GTK_WIN_POS_MOUSE);

		connect_to_method (d_fileSelection->get_ok_button()->clicked, 
					this, &fileSelectionOK);
		connect_to_method (d_fileSelection->get_cancel_button()->clicked,
					this, &fileSelectionCancel);
		}

	d_fileSelection->show ();
}


void TFGui::activateMenus (bool activate)
{
	Gtk_Widget	**w;
	int		c=0, i; 

	if (SanityCheck::warning((!this->d_menuHash->get_size()), "size==0", "TFGui::activateMenus"))
		return;
	
	w = new Gtk_Widget*[this->d_menuHash->get_size()];

	if (!activate || (d_HF && d_HF->isLoadedFromFile()))
		w[c++] = (Gtk_Widget *)this->d_menuHash->lookup (MENU_FILE_RELOAD);
	w[c++] = (Gtk_Widget *)this->d_menuHash->lookup (MENU_FILE_SAVE);
	w[c++] = (Gtk_Widget *)this->d_menuHash->lookup (MENU_FILE_SAVEAS);
	w[c++] = (Gtk_Widget *)this->d_menuHash->lookup (MENU_FILE_CLOSE);
	w[c++] = (Gtk_Widget *)this->d_menuHash->lookup (MENU_FILE_OPTIONS);
	if (activate)
		w[c++] = (Gtk_Widget *)this->d_menuHash->lookup (MENU_FILE_ABOUT);
	//w[c++] = (Gtk_Widget *)this->d_menuHash->lookup (MENU_HF_BRANCH);
	w[c++] = (Gtk_Widget *)this->d_menuHash->lookup (MENU_HF_FILL);
	w[c++] = (Gtk_Widget *)this->d_menuHash->lookup (MENU_HF_FLOWMAP);
	w[c++] = (Gtk_Widget *)this->d_menuHash->lookup (MENU_HF_FOLD);
	w[c++] = (Gtk_Widget *)this->d_menuHash->lookup (MENU_HF_INVERT);
	w[c++] = (Gtk_Widget *)this->d_menuHash->lookup (MENU_HF_LINSCALE);
	w[c++] = (Gtk_Widget *)this->d_menuHash->lookup (MENU_HF_MIRROR);
	w[c++] = (Gtk_Widget *)this->d_menuHash->lookup (MENU_HF_RESCALE);
	w[c++] = (Gtk_Widget *)this->d_menuHash->lookup (MENU_HF_ROTATE);
	w[c++] = (Gtk_Widget *)this->d_menuHash->lookup (MENU_HF_ROUGHEN);
	w[c++] = (Gtk_Widget *)this->d_menuHash->lookup (MENU_HF_SMOOTH);
	w[c++] = (Gtk_Widget *)this->d_menuHash->lookup (MENU_HF_TRANSFORM);
	w[c++] = (Gtk_Widget *)this->d_menuHash->lookup (MENU_HF_ERODE);
	w[c++] = (Gtk_Widget *)this->d_menuHash->lookup (MENU_HF_PLACEPOV);
	//w[c++] = (Gtk_Widget *)this->d_menuHash->lookup (MENU_VIEW_BRANCH);
	w[c++] = (Gtk_Widget *)this->d_menuHash->lookup (MENU_VIEW_2DPLANE);
	w[c++] = (Gtk_Widget *)this->d_menuHash->lookup (MENU_VIEW_3DWIRE);
	w[c++] = (Gtk_Widget *)this->d_menuHash->lookup (MENU_VIEW_3DHEIGHT);
	w[c++] = (Gtk_Widget *)this->d_menuHash->lookup (MENU_VIEW_3DLIGHT);
	w[c++] = (Gtk_Widget *)this->d_menuHash->lookup (MENU_VIEW_3DLIGHT);
	w[c++] = (Gtk_Widget *)this->d_menuHash->lookup (MENU_VIEW_FASTWIRE);
	if (!activate)
		w[c++] = (Gtk_Widget *)this->d_menuHash->lookup (MENU_VIEW_AUTOROTATE);
	//w[c++] = (Gtk_Widget *)this->d_menuHash->lookup (MENU_CMAP_BRANCH);
	w[c++] = (Gtk_Widget *)this->d_menuHash->lookup (MENU_CMAP_COLORBANDS);
	w[c++] = (Gtk_Widget *)this->d_menuHash->lookup (MENU_CMAP_DESERT);
	w[c++] = (Gtk_Widget *)this->d_menuHash->lookup (MENU_CMAP_GRAYSCALE);
	w[c++] = (Gtk_Widget *)this->d_menuHash->lookup (MENU_CMAP_REDHOT);

	for (i=0; i<c; i++)
		{
		if (!w[i])
			printf ("activateMenu Error: w[%d]=NULL\n", i);
		else
			w[i]->set_sensitive (activate);
		}

	delete w;
}


//
// Overloaded event implementations...
//
gint TFGui::button_press_event_impl (GdkEventButton *e)
{
	// button 1 starts rotating 
	if (e->button == 1)
		{
		d_rotate = TRUE;
		d_rotStartPosx = (int)e->x;	// Note current x position
		d_rotStartPosy = (int)e->y;	// Note current y position
		}
	else
	// button 2 resets the view 
	if (e->button == 2)
		{
		d_rotate = FALSE;
		if (d_HFD)
			{
			d_HFD->resetView ();
			d_HFD->draw (TRUE);
			}
		}

	return Gtk_Widget::button_press_event_impl(e);
}


gint TFGui::button_release_event_impl (GdkEventButton *e)
{
	//printf ("Button Release\n"); fflush (stdout);
	// Rotation completed
	d_rotate = FALSE;

	return Gtk_Widget::button_release_event_impl(e);
}


gint TFGui::motion_notify_event_impl (GdkEventMotion *e)
{
	int 		x, y;
	GdkModifierType	state;

	//printf ("Motion Notify\n"); fflush (stdout);

	if (d_rotate && d_fastWire) 
		{
		if (e->is_hint) 
			gdk_window_get_pointer(e->window, &x, &y, &state);
		else 
			{
			x = int (e->x);
			y = int (e->y);
			//state = e->state;
			}

		//printf ("Motion Notify (%d, %d)\n", x, y); fflush (stdout);
		// Apply rotation
		d_HFD->rotateFastView (0.0, (float)((d_rotStartPosx-x)/500.0), 0.0);
		if (d_HFD->d_fastDoRotZ)
			d_HFD->rotateFastView (0.0, 0.0, (float)((y-d_rotStartPosy)/500.0));
		d_rotStartPosx = int(e->x);
		d_rotStartPosy = int(e->y);
		d_HFD->draw (FALSE);
		}
	return Gtk_Widget::motion_notify_event_impl(e);
}


