/* Terraform - (C) 1997-1998 Robert Gasch (Robert_Gasch@peoplesoft.com)
 *  - www.peoplesoft.com/peoplepages/g/robert_gasch/terraform/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "terraform.h"

#include "math.h"
#include "TFGui.h"
#include "GlobalSanityCheck.h"
#include "GlobalTrace.h"
#include "GuiDialogOAC.h"
#include "HeightField.h"
#include "HeightFieldGenRandom.h"
#include "HeightFieldIO.h"
#include "MathGauss.h"

//
// instantiate static data members
//
int GlobalTrace::s_traceLevel=0;
const int GuiDialogOAC::s_HBOff=3;
const int GuiDialogOAC::s_VBOff=3;
const int MathGauss::s_gaussNRand=4;
const double MathGauss::s_maxRand=pow(2.0, 15.0)-1.0;
double MathGauss::s_gaussAdd=0.0;
double MathGauss::s_gaussFac=0.0;

int readhf (HeightField **HF, char *fname);



/*
 *  readhf: create a new HF object and then read into it
 */
int readhf (HeightField **HF, char *fname)
{
	int		rc=0;
	char		*ext;
	HeightFieldIO	*HFIO=NULL;

	SanityCheck::bailout ((*HF), "*HF!=NULL", "readhf()");
	SanityCheck::bailout ((!fname), "fname==NULL", "readhf()");
	SanityCheck::bailout ((!strlen(fname)), "strlen(fname)==0", "readhf()");

	*HF = new HeightField ();
	HFIO = new HeightFieldIO (fname);

	ext = HFIO->getExtension();
	if (!strcmp(ext,"png") || !strcmp(ext,"tiff") || !strcmp(ext,"tif"))
		{
		delete HFIO;
		printf ("Sorry -- PNG and TIFF read not supported from command line\n");
		delete *HF;
		*HF = NULL;
		return 1;
		}

	rc = HFIO->read (*HF);
	if (rc)
		delete *HF;
	HFIO->close ();
	//delete HFIO;		// XXX this causes a segmentation fault (why??)

	return (rc);
}


/*
 * help: display recognized command line options 
 */
static void help ()
{
	printf ("terraform %d.%d.%d\n", TF_MAJOR_VERSION, TF_MINOR_VERSION, 
		TF_MICRO_VERSION);
	printf ("  -dbg, --debug <level>		set debug to the specified level\n");
	printf ("  -f, --file <file>		read Height Field <file> upon startup\n");
	printf ("  -g, --generate 		generate a Height Field upon startup\n");
	printf ("  -version, --version		print version\n");
	printf ("  -v, --verbose			verbose mode\n");
	printf ("  -?, --help			print help and exit\n");
	printf ("\n");
}


/*
 *  process_app_args: process the application-related command line arguments 
 */
static void process_app_args (HeightField **HF,int argc, char **argv)
{
	// int			do_gui=TRUE;
	int			rc;

	for (int i=1; i<argc; i++)
		{
		if (argv[i] &&
	            (!strcmp(argv[i],"-f")||!strcmp(argv[i],"--file")) )
			{
			if (argc >= i+1 && argv[i+1])
				{
				if ((rc=readhf (HF, argv[++i])) < 0)
					{
					printf ("File read failed (rc=%d) for file [%s] ... Exiting\n", rc, argv[i]);
					exit (-1);
					}
				}
			else
				{
				printf ("Error: file argument for %s <file> missing ... Exiting\n", argv[i]);
				exit (-1);
				}
			}
		else
		if (!strcmp (argv[i], "-g") || !strcmp (argv[i], "--generate"))
			{
			*HF = new HeightField ();
			HeightFieldGenRandom::generate (*HF, 400);
			}
#ifdef NEVER
		else
		if (!strcmp (argv[i], "-cli"))
			do_gui=FALSE;
#endif
		else 

		// dummy checks we do in process_global_args
		if (!strcmp (argv[i], "-dbg") || !strcmp (argv[i], "--debug"))
			i++;
		else
		if (!strcmp (argv[i], "-v") || !strcmp (argv[i], "--verbose"))
			;
		else
			{
			printf ("Error: argument [%s] not recognized ... \n", 
				argv[i]);
			help ();
			exit (0);
			}
		}
} 


/*
 *  process_global_args: process the global command line arguments 
 */
static void process_global_args (int argc, char **argv)
{
	// see if we should set a certain trace level 
	for (int i=1; i<argc; i++)
		{
		if (!strcmp (argv[i], "-dbg") || !strcmp (argv[i], "--debug"))
			{
			if (argc >= i+2)	// 1 for count-to-index, 1 for next arg
				GlobalTrace::setTraceLevel (atoi (argv[++i]));
			else
				{
				printf ("Error: traceLevel argument missing for %s ... Exiting\n", argv[i]);
				exit (1);
				}
			}
		else
		if (!strcmp (argv[i], "-v") || !strcmp (argv[i], "--verbose"))
			GlobalTrace::setTraceLevel (GlobalTrace::TRACE_VERBOSE);
		else
		if (!strcmp (argv[i], "-version") || !strcmp(argv[i], "--version"))
			{
			printf ("terraform %d.%d.%d\n", TF_MAJOR_VERSION, 
				TF_MINOR_VERSION, TF_MICRO_VERSION);
			exit (0);
			}
		else
		if (!strcmp (argv[i], "--help") || !strcmp (argv[i], "-?"))
			{
			help ();
			exit (0);
			}

		}
}


/*
 *  main: run the terraform program 
 */
int main (int argc, char *argv[])
{
	int		do_gui=TRUE;	//
	HeightField	*HF=NULL;

	// process command line arguments 
	process_global_args (argc, argv);
	process_app_args (&HF, argc, argv);

	// use graphical user interface
	if (do_gui)
		{
		Gtk_Main	gtkMain (&argc, &argv);
		TFGui	*tfGui;

		tfGui = new TFGui (&gtkMain, argc, argv);
		tfGui->show ();
		if (HF)
			tfGui->setHeightField (HF);
		gtkMain.run();
		}
/*
	// use command line interface
	else
		{
		if (!(HFP = new HeightFieldParser (verbose)))
			{
			cerr << "Error instantiating HeightFieldParser ... Exiting\n";
			exit (NOGOOD);
			}
		HFP->P_Parse ();
		delete HFP;
		exit (1);
		}
*/

	exit (0);
}
